from __future__ import annotations

import hashlib
import os
import shutil
import tempfile
import threading
import warnings
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import uxarray as uxr
import xarray as xr
from numpy.typing import ArrayLike, NDArray
from scipy.optimize import OptimizeWarning, curve_fit
from uxarray import INT_FILL_VALUE, UxDataArray, UxDataset

from cratermaker._cratermaker import surface_functions
from cratermaker.constants import (
    _COMBINED_DATA_FILE_NAME,
    _GRID_FILE_NAME,
    _SMALLFAC,
    _SURFACE_DIR,
    _VSMALL,
    FloatLike,
)
from cratermaker.utils import export
from cratermaker.utils.component_utils import ComponentBase, import_components
from cratermaker.utils.general_utils import (
    format_large_units,
    validate_and_normalize_location,
)
from cratermaker.utils.montecarlo_utils import get_random_location_on_face

if TYPE_CHECKING:
    from uxarray.grid.neighbors import SpatialHash

    from cratermaker.components.target import Target

surface_lock = threading.Lock()


class Surface(ComponentBase):
    _registry: dict[str, type[Surface]] = {}

    """
    Used for handling surface-related data and operations in the cratermaker project.

    It provides methods for setting elevation data, calculating distances and bearings, and other surface-related computations.
    The Surface class extends UxDataset for the cratermaker project.

    Parameters
    ----------
    target : Target, optional
        The target body or name of a known target body for the impact simulation.
    reset : bool, optional
        Flag to indicate whether to reset the surface. Default is the value of `regrid`
    regrid : bool, optional
        Flag to indicate whether to regrid the surface. Default is False.
    simdir : str | Path
        The main project simulation directory. Defaults to the current working directory if None.
    **kwargs : Any
        Additional keyword arguments.
    """

    def __init__(
        self,
        target: Target | str | None = None,
        simdir: str | Path | None = None,
        **kwargs,
    ):
        from cratermaker.components.target import Target

        object.__setattr__(self, "_target", None)
        object.__setattr__(self, "_uxds", None)
        object.__setattr__(self, "_pix", None)
        object.__setattr__(self, "_pix_mean", None)
        object.__setattr__(self, "_pix_std", None)
        object.__setattr__(self, "_pix_min", None)
        object.__setattr__(self, "_pix_max", None)
        object.__setattr__(self, "_area", None)
        object.__setattr__(self, "_edge_tree", None)
        object.__setattr__(self, "_edge_face_distance", None)
        object.__setattr__(self, "_edge_lengths", None)
        object.__setattr__(self, "_edge_indices", None)
        object.__setattr__(self, "_node_tree", None)
        object.__setattr__(self, "_face_tree", None)
        object.__setattr__(self, "_face_area", None)
        object.__setattr__(self, "_face_size", None)
        object.__setattr__(self, "_face_bin_indices", None)
        object.__setattr__(self, "_face_bin_argmin", None)
        object.__setattr__(self, "_face_bin_argmax", None)
        object.__setattr__(self, "_face_bin_area", None)
        object.__setattr__(self, "_face_x", None)
        object.__setattr__(self, "_face_y", None)
        object.__setattr__(self, "_face_z", None)
        object.__setattr__(self, "_node_x", None)
        object.__setattr__(self, "_node_y", None)
        object.__setattr__(self, "_node_z", None)
        object.__setattr__(self, "_smallest_length", None)

        super().__init__(simdir=simdir, **kwargs)

        self._data_variable_init = {
            "node_elevation": {
                "units": "m",
                "long_name": "elevation of nodes",
                "initial_value": 0.0,
                "isfacedata": False,
            },
            "face_elevation": {
                "units": "m",
                "long_name": "elevation of faces",
                "initial_value": 0.0,
                "isfacedata": True,
            },
        }

        self.target = Target.maker(target, **kwargs)

        return

    def __str__(self) -> str:
        base = super().__str__()
        return (
            f"{base}\nTarget: {self.target.name}\nGrid File: {self.grid_file}\n"
            f"Number of faces: {self.n_face}\n"
            f"Number of nodes: {self.n_node}"
        )

    def __del__(self):
        try:
            if hasattr(self, "_uxds") and hasattr(self._uxds, "close"):
                if hasattr(self._uxds, "uxgrid") and hasattr(self._uxds.uxgrid, "_ds"):
                    self._uxds.uxgrid._ds.close()
                self._uxds.close()
        except Exception as e:
            warnings.warn(f"An error occurred while closing the dataset: {e}", RuntimeWarning, stacklevel=2)
        try:
            if hasattr(self, "_grid") and hasattr(self._grid, "uxgrid") and hasattr(self._grid.uxgrid, "_ds"):
                self._grid.uxgrid._ds.close()
        except Exception as e:
            warnings.warn(f"An error occurred while closing the dataset: {e}", RuntimeWarning, stacklevel=2)

    @classmethod
    def maker(
        cls: Surface,
        surface: str | Surface | None = None,
        target: Target | str | None = None,
        reset: bool = False,
        regrid: bool = False,
        simdir: str | Path | None = None,
        **kwargs,
    ) -> Surface:
        """
        Factory method to create a Surface instance from a grid file.

        Parameters
        ----------
        surface : str or Surface, optional
            The name of the type of grid used for the surface. Default is "icosphere".
        target : Target, optional
            The target body or name of a known target body for the impact simulation.
        reset : bool, optional
            Flag to indicate whether to reset the surface. Default is True.
        regrid : bool, optional
            Flag to indicate whether to regrid the surface. Default is False.
        simdir : str | Path
            The main project simulation directory. Defaults to the current working directory if None.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Surface
            An initialized Surface object.
        """
        if surface is None:
            surface = "icosphere"

        surface = super().maker(
            component=surface,
            target=target,
            reset=reset,
            regrid=regrid,
            simdir=simdir,
            **kwargs,
        )
        return surface

    def reset(self, **kwargs: Any) -> None:
        """
        Reset the surface to its initial state.
        """
        # Remove all old data from the dataset
        varlist = list(self.uxds.data_vars)
        for name in varlist:
            if name not in self._data_variable_init:
                del self.uxds[name]
        for name, entry in self._data_variable_init.items():
            self._add_new_data(
                name=name,
                data=entry["initial_value"],
                long_name=entry["long_name"],
                units=entry["units"],
                isfacedata=entry["isfacedata"],
                save_to_file=True,
            )

        return

    def extract_region(self, location: tuple[FloatLike, FloatLike], region_radius: FloatLike):
        """
        Extract a regional grid based on a given location and radius.

        Parameters
        ----------
        location : tuple[float, float]
            tuple containing the longitude and latitude of the location in degrees.
        region_radius : float
            The radius of the region to extract in meters.

        Returns
        -------
        LocalSurface
            A LocalSurface object containing a view of the regional grid.

        """
        region_angle = np.rad2deg(region_radius / self.radius)
        if len(location) == 1:
            location = location.item()
        location = validate_and_normalize_location(location)
        coords = np.asarray(location)

        if region_angle < 180.0:
            face_indices = self.face_tree.query_radius(coords, region_angle)
            if len(face_indices) == 0:
                return None

            # First select edges and nodes that are attached to these faces
            edge_indices = np.unique(self.face_edge_connectivity[face_indices].ravel())
            edge_indices = edge_indices[edge_indices != INT_FILL_VALUE]

            node_indices = np.unique(self.face_node_connectivity[face_indices].ravel())
            node_indices = node_indices[node_indices != INT_FILL_VALUE]

            # Now add in all faces that are connected to anything inside the region, so that the outermost border of the local region has a buffer of faces
            # These are needed for diffusion calculations
            neighbor_faces = self.face_face_connectivity[face_indices]
            neighbor_faces = neighbor_faces[neighbor_faces != INT_FILL_VALUE]
            node_faces = self.node_face_connectivity[node_indices]
            node_faces = node_faces[node_faces != INT_FILL_VALUE]
            edge_faces = self.edge_face_connectivity[edge_indices]
            edge_faces = edge_faces[edge_faces != INT_FILL_VALUE]
            face_indices = np.unique(np.concatenate((face_indices, neighbor_faces, node_faces, edge_faces)))
        else:  # This is the entire surface
            face_indices = slice(None)
            edge_indices = slice(None)
            node_indices = slice(None)

        return LocalSurface(
            surface=self,
            face_indices=face_indices,
            node_indices=node_indices,
            edge_indices=edge_indices,
            location=location,
            region_radius=region_radius,
        )

    def add_data(
        self,
        name: str,
        data: FloatLike | NDArray,
        long_name: str | None = None,
        units: str | None = None,
        isfacedata: bool = True,
        overwrite: bool = False,
    ) -> None:
        """
        Adds new data to the surface.

        If the data variable already exists, it will be overwritten if `overwrite` is set to True.

        Parameters
        ----------
        name : str
            Name of the data variable. This will also be used as the data file name.
        data : scalar or array-like
            Data file to be saved. If data is a scalar, then the data file will be filled with that value. If data is an array, then the data file will be filled with the array values. The data array must have the same size as the number of faces or nodes in the grid.
        long_name : str, optional
            Long name of the data variable that will be saved as an attribute if this is new data. If the data already exists on the surface, this will be ignored.
        units : str, optional
            Units of the data variable that will be saved as an attribute if this is new data. If the data already exists on the surface, this will be ignored.
        isfacedata : bool, optional, default True
            Flag to indicate whether the data is face data or node data. This is only needed if `data` is a scalar, otherwise it is ignored
        overwrite : bool, optional, default False
            By default, new data is added to the old data. This flag indicates that the data should be overwritten, replacing any old data with the new data.


        Returns
        -------
        None
        """
        return self._full().add_data(
            name=name,
            data=data,
            long_name=long_name,
            units=units,
            isfacedata=isfacedata,
            overwrite=overwrite,
        )

    def update_elevation(
        self,
        new_elevation: ArrayLike | FloatLike,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Update the elevation data for the target's surface mesh. This method will determine whether to update the node or face data (or both) depending on the size of the input data. If a scalar is passed, both node and face elevation will be updated to that value. By default, the elevation data will be added to the existing data. If `overwrite` is set to True, the existing data will be replaced with the new data.

        Parameters
        ----------
        new_elevation : ArrayLike | FloatLike
            Elevation to be added (or replaced, if overwrite is True). This can be a scalar, an array with the same size as the number of faces, an array with the same size as the number of nodes, or an array with the same size as the number of faces + nodes.
        overwrite : bool, optional
            If True, the existing data will be replaced with the new data. Default is False.
        **kwargs : Any
            Additional keyword arguments.

        Notes
        -----
        When passing combined data, the first part of the array will be used for face elevation and the second part for node elevation.
        """
        return self._full().update_elevation(new_elevation=new_elevation, overwrite=overwrite, **kwargs)

    def apply_diffusion(self, kdiff: FloatLike | NDArray) -> None:
        """
        Apply diffusion to the surface.

        Parameters
        ----------
        kdiff : float or array-like
            The degradation state of the surface, which is the product of diffusivity and time. It can be a scalar or an array of the same size as the number of faces in the grid.
            If it is a scalar, the same value is applied to all faces. If it is an array, it must have the same size as the number of faces in the grid.
            The value of kdiff must be greater than 0.0.

        """
        return self._full().apply_diffusion(kdiff)

    def slope_collapse(self, critical_slope_angle: FloatLike = 35.0) -> None:
        """
        Collapse all slopes larger than the critical slope angle.

        Parameters
        ----------
        critical_slope_angle : float
            The critical slope angle (angle of repose) in degrees.
        """
        return self._full().slope_collapse(critical_slope_angle)

    def compute_slope(self) -> NDArray[np.float64]:
        """
        Compute the slope of the surface.

        Returns
        -------
        NDArray[np.float64]
            The slope of all faces in degrees.
        """
        return self._full().compute_slope()

    def apply_noise(
        self,
        model: str = "turbulence",
        noise_width: FloatLike = 1000e3,
        noise_height: FloatLike = 1e3,
        **kwargs: Any,
    ) -> None:
        """
        Apply noise to the surface.

        Parameters
        ----------
        model : str
            The noise model to use. Options are "turbulence"
        noise_width : float
            The width of the noise in meters.
        noise_height : float
            The height of the noise in meters.
        kwargs : Any
            Additional arguments to pass to the noise model.
        """
        return self._full().apply_noise(model=model, noise_width=noise_width, noise_height=noise_height, **kwargs)

    def calculate_face_and_node_distances(self, location: tuple[float, float]) -> tuple[NDArray, NDArray]:
        """
        Computes the distances from a given location to all faces and nodes.

        Parameters
        ----------
        location : tuple[float, float]
            tuple containing the longitude and latitude of the location in degrees.

        Returns
        -------
        NDArray
            Array of distances for each face in meters.
        NDArray
            Array of distances for each node in meters.
        """
        return self._full().calculate_face_and_node_distances(location)

    def calculate_face_and_node_bearings(self, location: tuple[float, float]) -> tuple[NDArray, NDArray]:
        """
        Computes the initial bearing from a given location to all faces and nodes.

        Parameters
        ----------
        location : tuple[float, float]
            tuple containing the longitude and latitude of the location in degrees.

        Returns
        -------
        NDArray
            Array of initial bearings for each face in radians.
        NDArray
            Array of initial bearings for each node in radians.

        Notes
        -----
        This is intended to be used as a helper to calculate_face_and_node_bearings.
        """
        return self._full().calculate_face_and_node_bearings(location)

    def find_nearest_node(self, location):
        """
        Find the index of the nearest node to a given point.

        This method calculates the Haversine distance from the given point to each node in the grid, and returns the index of the node with the minimum distance.

        Parameters
        ----------
        location : tuple
            A tuple containing two elements: (longitude, latitude) in degrees.

        Returns
        -------
        int
            The index of the nearest node in the grid to the given point.

        Notes
        -----
        The method uses the ball tree query method that is included in the UxArray.Grid class.
        """
        location = validate_and_normalize_location(location)
        if len(location) == 1:
            location = location.item()
        coords = np.asarray(location)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", Warning)
            node_ind = self.node_tree.query(coords=coords, k=1, return_distance=False)

        return node_ind.item()

    def find_nearest_face(self, location):
        """
        Find the index of the nearest face to a given point.

        This method calculates the Haversine distance from the given point to each face in the grid, and returns the index of the face with the minimum distance.

        Parameters
        ----------
        location : tuple
            A tuple containing two elements: (longitude, latitude) in degrees.

        Returns
        -------
        int
            The index of the nearest face in the grid to the given point.

        Notes
        -----
        The method uses the ball tree query method that is included in the UxArray.Grid class.  It is only approximate, as it looks for whichever face center is closest to the location. This means that it can return a neighboring face, rather than the face that contains the point.
        """
        location = validate_and_normalize_location(location)
        if len(location) == 1:
            location = location.item()
        coords = np.asarray(location)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", OptimizeWarning)
            face_ind = self.face_tree.query(coords=coords, k=1, return_distance=False)

        return face_ind.item()

    def interpolate_node_elevation_from_faces(self) -> None:
        """
        Update node elevations by area-weighted averaging of adjacent face elevations.

        For each node, the elevation is computed as the area-weighted average of the elevations
        of the surrounding faces.

        Returns
        -------
        None
        """
        face_elevation = self.face_elevation
        face_area = self.face_area
        node_face_conn = self.node_face_connectivity

        node_elevation = np.zeros(self.n_node, dtype=np.float64)

        for node_id in range(self.n_node):
            connected_faces = node_face_conn[node_id]
            valid = connected_faces != INT_FILL_VALUE
            faces = connected_faces[valid]

            if faces.size == 0:
                continue

            areas = face_area[faces]
            elevations = face_elevation[faces]

            total_area = np.sum(areas)
            if total_area > 0:
                node_elevation[node_id] = np.sum(elevations * areas) / total_area

        self.node_elevation = node_elevation

    def get_random_location_on_face(self, face_index: int, **kwargs) -> float | tuple[float, float] | ArrayLike:
        """
        Generate a random coordinate within a given face of a the mesh.

        Parameters
        ----------
        grid : uxarray.Grid
            The grid object containing the mesh information.
        face_index : int | NDArray[np.int64]
            The index or array of indices of the face within the grid to obtain the random sample.
        size : int or tuple of ints, optional
            The number of samples to generate. If size is None (the default), a single tuple is returned. If size is greater than 1,
            then an array of tuples is returned. The default is 1.

        Returns
        -------
        (lon,lat) or ndarray[(lon,lat)] of given size
            A pair or array of pairs of longitude and latitude values in degrees.

        Notes
        -----
        This method is a wrapper for :func:`cratermaker.utils.montecarlo_utils.get_random_location_on_face`.
        """
        return get_random_location_on_face(self.uxgrid, face_index, rng=self.rng, **kwargs)

    def elevation_to_cartesian(self, element="face") -> NDArray[np.float64]:
        """
        Convert either the face or node elevations to Cartesian coordinates.

        Parameters
        ----------
        element : str, optional
            The type of element to convert. Can be "face" or "node". Default is "face".

        Returns
        -------
        NDArray[np.float64, 3]
            The Cartesian coordinates of the face elevations.
        """
        return self._full().elevation_to_cartesian(element=element)

    def save(
        self,
        combine_data_files: bool = False,
        interval_number: int = 0,
        time_variables: dict | None = None,
        **kwargs,
    ) -> None:
        """
        Save the surface data to the specified directory. Each data variable is saved to a separate NetCDF file. If 'time_variables' is specified, then a one or more variables will be added to the dataset along the time dimension. If 'interval_number' is included as a key in `time_variables`, then this will be appended to the data file name.

        Parameters
        ----------
        combine_data_files : bool, optional
            If True, combine all data variables into a single NetCDF file, otherwise each variable will be saved to its own NetCDF file. Default is False.
        interval_number : int, optional
            Interval number to append to the data file name. Default is 0.
        time_variables : dict, optional
            Dictionary containing one or more variable name and value pairs. These will be added to the dataset along the time dimension. Default is None.
        """
        do_not_save = ["face_area"]

        self.data_dir.mkdir(parents=True, exist_ok=True)

        if time_variables is None:
            time_variables = {"elapsed_time": float(interval_number)}
        else:
            if not isinstance(time_variables, dict):
                raise TypeError("time_variables must be a dictionary")

        self.uxds.close()

        ds = self.uxds.expand_dims(dim="time").assign_coords({"time": [interval_number]})
        for k, v in time_variables.items():
            ds[k] = xr.DataArray(data=[v], name=k, dims=["time"], coords={"time": [interval_number]})

        drop_vars = [k for k in ds.data_vars if k in do_not_save]
        if len(drop_vars) > 0:
            ds = ds.drop_vars(drop_vars)

        self._save_data(ds, interval_number, combine_data_files)

        save_geometry = interval_number == 0
        self.export(
            format="vtp", interval_number=interval_number, time_variables=time_variables, save_geometry=save_geometry, **kwargs
        )

        return

    def export(self, format="vtp", **kwargs) -> None:
        """
        Export the surface mesh to a file in the specified format. Currently only VTK is supported.
        """
        if format == "vtp" or format == "vtk":
            export.to_vtk(self, **kwargs)
        else:
            raise ValueError(f"Unsupported export format: {format}")

    @staticmethod
    def _sphere_function(coords, x_c, y_c, z_c, r):
        """
        Compute the sphere function.

        Parameters
        ----------
        coords : ndarray
            Array of x, y, and z coordinates.
        x_c, y_c, z_c : float
            Center coordinates of the sphere.
        r : float
            Radius of the sphere.

        Returns
        -------
        ndarray
            The values of the sphere function at the given coordinates.
        """
        x, y, z = coords.T
        return (x - x_c) ** 2 + (y - y_c) ** 2 + (z - z_c) ** 2 - r**2

    def _calculate_distance(
        self,
        lon1: FloatLike,
        lat1: FloatLike,
        lon2: ArrayLike,
        lat2: ArrayLike,
    ) -> NDArray[np.float64]:
        """
        Calculate the great circle distance between one point and one or more other points in meters.

        Parameters
        ----------
        lon1 : FloatLike
            Longitude of the first point in radians.
        lat1 : FloatLike
            Latitude of the first point in radians.
        lon2 : FloatLike or ArrayLike
            Longitude of the second point or array of points in radians.
        lat2 : FloatLike or ArrayLike
            Latitude of the second point or array of points in radians.

        Returns
        -------
        NDArray
            Great circle distance between the two points in meters.

        Notes
        -----
        This is a wrapper for a compiled Rust function and is intended to be used as a helper to calculate_face_and_node_distances.
        """
        # Validate that lon1 and lat1 are single points
        if not np.isscalar(lon1):
            if lon1.size != 1:
                raise ValueError("lon1 must be a single point")
            lon1 = lon1.item()
        if not np.isscalar(lat1):
            if lat1.size != 1:
                raise ValueError("lat1 must be a single point")
            lat1 = lat1.item()
        if np.isscalar(lon2):
            lon2 = np.array([lon2])
        if np.isscalar(lat2):
            lat2 = np.array([lat2])

        return surface_functions.calculate_distance(lon1=lon1, lat1=lat1, lon2=lon2, lat2=lat2, radius=self.radius)

    @staticmethod
    def _calculate_bearing(
        lon1: FloatLike,
        lat1: FloatLike,
        lon2: FloatLike | ArrayLike,
        lat2: FloatLike | ArrayLike,
    ) -> NDArray[np.float64]:
        """
        Calculate the initial bearing from one point to one or more other points in radians.

        Parameters
        ----------
        lon1 : FloatLike
            Longitude of the first point in radians.
        lat1 : FloatLike
            Latitude of the first point in radians.
        lon2 : FloatLike or ArrayLike
            Longitude of the second point or array of points in radians.
        lat2 : FloatLike or ArrayLike
            Latitude of the second point or array of points in radians.

        Returns
        -------
        NDArray
            Initial bearing from the first point to the second point or points in radians.
        """
        return LocalSurface._calculate_bearing(lon1=lon1, lat1=lat1, lon2=lon2, lat2=lat2)

    @staticmethod
    def _compute_elevation_to_cartesian(position: NDArray, elevation: NDArray) -> NDArray:
        """
        Convert elevation values to Cartesian coordinates.

        Parameters
        ----------
        position : NDArray
            The position of the points in Cartesian coordinates.
        elevation : NDArray
            The elevation values to convert.

        Returns
        -------
        NDArray
            The Cartesian coordinates of the points with the given elevation.
        """
        runit = position / np.linalg.norm(position, axis=1, keepdims=True)

        return position + elevation[:, np.newaxis] * runit

    def _load_from_files(self, reset: bool = False, **kwargs: Any) -> None:
        """
        Load the grid and data files into the surface object.

        This function loads the grid file and data files from the specified directory. If the grid file does not exist, it will attempt to create a new grid.
        If the data files do not exist, it will create an empty dataset. If reset is True, it will delete all data files except the grid file.

        Parameters
        ----------
        reset : bool, optional
            Flag to indicate whether to reset the surface. Default is False.
        """
        # Get the names of all data files in the data directory that are not the grid file
        regrid = self._regrid_if_needed(**kwargs)
        reset = reset or regrid

        data_file_list = list(self.data_dir.glob("*.nc"))
        if self.grid_file in data_file_list:
            data_file_list.remove(self.grid_file)

        # if data_file_list is empty, set reset to True
        reset = reset or not data_file_list

        # If reset is True, delete all data files except the grid file
        if reset:
            for f in data_file_list:
                f.unlink()
            data_file_list = []

        try:
            with xr.open_dataset(self.grid_file) as uxgrid:
                uxgrid.load()
                if reset:  # Create an empty dataset
                    self._uxds = uxr.UxDataset()
                else:  # Read data from from existing datafiles
                    with uxr.open_mfdataset(uxgrid, data_file_list, use_dual=False) as ds:
                        self._uxds = ds.isel(time=-1).load()
                self._uxds.uxgrid = uxr.Grid.from_dataset(uxgrid)
                self._uxgrid = uxgrid
        except Exception as e:
            raise RuntimeError("Error loading grid and data files") from e

        if reset:
            self.reset(**kwargs)

        return

    def _generate_grid(self, **kwargs: Any) -> None:
        """
        Generate a tessellated mesh of a sphere of based on the particular Surface component that is being used.
        """
        self.grid_file.unlink(missing_ok=True)

        points = self._generate_face_distribution(**kwargs)

        threshold = min(10 ** np.floor(np.log10(self.pix / self.radius)), 1e-6)
        uxgrid = uxr.Grid.from_points(points, method="spherical_voronoi", threshold=threshold)
        uxgrid.attrs["_id"] = self._id
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            uxgrid.to_xarray().to_netcdf(temp_file.name)
            temp_file.flush()
            os.fsync(temp_file.fileno())

        # Replace the original file only if writing succeeded
        shutil.move(temp_file.name, self.grid_file)
        self._uxgrid = uxgrid

        regrid = self._regrid_if_needed(**kwargs)
        if regrid:
            raise ValueError("Grid file does not match the expected parameters.")
        self._compute_face_size(uxgrid)

        return

    def _regrid_if_needed(self, regrid: bool = False, **kwargs: Any) -> bool:
        """
        Check if the existing grid matches the desired parameters determine if regridding is necessary.

        This function checks if a grid file exists and matches the specified parameters based on a unique hash generated from these
        parameters. If the grid does not exist or does not match the parameters it generates a new grid and returns True.

        Parameters
        ----------
        regrid: bool, optional
            Flag to force regridding even if the grid file exists. Default is False.

        Returns
        -------
        bool
            A boolean indicating whether the grid should be regenerated.
        """
        # Find out if the file exists, if it does't we'll need to make a new grid
        self.data_dir.mkdir(parents=True, exist_ok=True)
        regrid = regrid or not Path(self.grid_file).exists()

        if not regrid:
            try:
                with xr.open_dataset(self.grid_file) as ds:
                    ds.load()
                    uxgrid = uxr.Grid.from_dataset(ds)
                    old_id = uxgrid.attrs.get("_id")
                    regrid = old_id != self._id
            except Exception:
                # Failed to open an old file for whatever reason, so we'll need to regrid
                regrid = True

        if regrid:
            print("Creating a new grid")
            try:
                self._generate_grid(**kwargs)
            except Exception as e:
                raise RuntimeError("Failed to create a new grid") from e

        return regrid

    def _full(self):
        return LocalSurface(self, face_indices=slice(None), node_indices=slice(None), edge_indices=slice(None))

    def _save_data(self, ds: xr.Dataset | xr.DataArray, interval_number: int = 0, combine_data_files: bool = False) -> None:
        """
        Save the data to the specified directory.

        If `combine_data_files` is True, then all data variables are saved to a single NetCDF
        file. If False, then only the data variables for the current interval are saved to a NetCDF file with the interval number
        appended.

        Parameters
        ----------
        ds : xr.Dataset or xr.DataArray
            The data to be saved.
        interval_number : int, Default is 0.
            Interval number to append to the data file name. Default is 0.
        combine_data_files : bool, Default is False.
            If True, combine all data variables into a single NetCDF file, otherwise each variable will be saved to its own NetCDF file.

        Notes
        -----
        This function first saves to a temporary file and then moves that file to the final destination. This is done to avoid file
        locking issues with NetCDF files.
        """
        if isinstance(ds, xr.DataArray):
            ds = ds.to_dataset()

        if "time" not in ds.dims:
            ds = ds.expand_dims(["time"])
        if "time" not in ds.coords:
            ds = ds.assign_coords({"time": [interval_number]})

        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            if combine_data_files:
                filename = _COMBINED_DATA_FILE_NAME
            else:
                filename = _COMBINED_DATA_FILE_NAME.replace(".nc", f"{interval_number:06d}.nc")

            data_file = self.data_dir / filename
            if data_file.exists():
                with xr.open_mfdataset(data_file) as ds_file:
                    ds_file = ds.merge(ds_file, compat="override")
                    ds_file.load()
            else:
                ds_file = ds

            temp_file = Path(temp_dir) / filename

            comp = {"zlib": True, "complevel": 9}
            encoding = dict.fromkeys(ds_file.data_vars, comp)
            ds_file.to_netcdf(temp_file, encoding=encoding)
            ds_file.close()
            shutil.move(temp_file, data_file)

        return

    def _add_new_data(
        self,
        name: str,
        long_name: str | None = None,
        units: str | None = None,
        data: FloatLike | NDArray | None = None,
        isfacedata: bool = True,
        save_to_file: bool = False,
        interval_number: int = 0,
        combine_data_files: bool = False,
    ) -> None:
        """
        Generate either a node or face data variable and optionally save it to a file. If the data variable already exists, it will be overwritten.

        Parameters
        ----------
        name : str
            Name of the data variable. This will also be used as the data file name.
        long_name : str, optional
            Long name of the data variable that will be saved as an attribute.
        units : str, optional
            Units of the data variable that will be saved as an attribute.
        data : scalar or array-like
            Data file to be saved. If data is a scalar, then the data file will be filled with that value. If data is an array, then the data file will be filled with the array values. The data array must have the same size as the number of faces or nodes in the grid.
        isfacedata : bool, optional
            Flag to indicate whether the data is face data or node data. Default is True.
        save_to_file: bool, optional
            Specify whether the data should be saved to a file. Default is False.
        interval_number : int, optional, default 0
            The interval number to use when saving the data to the data file.
        combine_data_files : bool, optional
            If True, combine the current data with the existing data for previous intervals in the data file. Default is False.

        Returns
        -------
        None
        """
        with xr.open_dataset(self.grid_file) as ds:
            ds.load()
            uxgrid = uxr.Grid.from_dataset(ds)
        if long_name is None and units is None:
            attrs = None
        else:
            attrs = {}
            if long_name:
                attrs["long_name"] = long_name
            if units:
                attrs["units"] = units

        if isfacedata:
            dims = ["n_face"]
            size = uxgrid.n_face
        else:
            dims = ["n_node"]
            size = uxgrid.n_node

        if data is None:
            data = np.zeros(size, dtype=np.float64)
        elif np.isscalar(data):
            data = np.full(size, data)
        else:
            if data.size != size:
                raise ValueError("data must have the same size as the number of faces or nodes in the grid")
        uxda = UxDataArray(
            data=data,
            dims=dims,
            attrs=attrs,
            name=name,
            uxgrid=uxgrid,
        )

        self._uxds[name] = uxda

        if save_to_file:
            self._save_data(uxda, interval_number, combine_data_files)
        return

    @abstractmethod
    def _generate_face_distribution(self, **kwargs: Any) -> tuple[NDArray, NDArray, NDArray]: ...

    def _compute_face_size(self, uxgrid: UxDataset | None = None) -> None:
        """
        Compute the effective pixel size of the mesh based on the face areas.

        Parameters
        ----------
        uxgrid : UxDataset
            The grid object containing the mesh information. If not set, it will be retrieved self.uxds.uxgrid

        """
        if uxgrid is None:
            if self.uxgrid is None:
                return
            else:
                uxgrid = self.uxgrid
        self._face_area = uxgrid.face_areas.values * self.radius**2
        self._face_size = np.sqrt(self._face_area)
        self._pix_mean = float(self._face_size.mean().item())
        self._pix_std = float(self._face_size.std().item())
        self._pix_min = float(self._face_size.min().item())
        self._pix_max = float(self._face_size.max().item())
        return

    @property
    def _hashvars(self):
        """
        The variables used to generate the hash.
        """
        return [self._component_name, self._radius]

    @property
    def _id(self):
        """
        The hash id of the grid. This is used for determining if the grid needs to be regridded.
        """
        combined = ":".join(str(v) for v in self._hashvars)
        hash_object = hashlib.sha256(combined.encode())
        return hash_object.hexdigest()

    @property
    def uxds(self) -> UxDataset:
        """
        The data associated with the surface. This is an instance of UxDataset.
        """
        return self._uxds

    @property
    def uxgrid(self):
        """
        The grid object.
        """
        if self.uxds is not None:
            return self.uxds.uxgrid

    @property
    def data_dir(self):
        """
        Directory for data files.
        """
        return self.simdir / _SURFACE_DIR

    @property
    def gridtype(self):
        """
        The name of the grid type.
        """
        return self._component_name

    @property
    def grid_file(self):
        """
        Path to the grid file.
        """
        return self.simdir / _SURFACE_DIR / _GRID_FILE_NAME

    @property
    def target(self):
        """
        The target body for the impact simulation. Set during initialization.
        """
        return self._target

    @target.setter
    def target(self, value):
        from cratermaker.components.target import Target

        self._target = Target.maker(value)
        return

    @property
    def face_elevation(self) -> NDArray[np.float64]:
        """
        The elevation of the faces.
        """
        return self.uxds["face_elevation"].values

    @face_elevation.setter
    def face_elevation(self, value: NDArray) -> None:
        """
        Set the elevation of the faces.

        Parameters
        ----------
        value : NDArray
            The elevation values to set for the faces.
        """
        value = np.asarray(value, dtype=np.float64)
        if value.size != self.n_face:
            raise ValueError(f"Value must have size {self.n_face}, got {value.size} instead.")
        self.uxds["face_elevation"][:] = value
        return

    @property
    def node_elevation(self) -> NDArray[np.float64]:
        """
        The elevation of the nodes.
        """
        return self.uxds["node_elevation"].values

    @node_elevation.setter
    def node_elevation(self, value: NDArray) -> None:
        """
        Set the elevation of the nodes.

        Parameters
        ----------
        value : NDArray
            The elevation values to set for the nodes.
        """
        value = np.asarray(value, dtype=np.float64)
        if value.size != self.n_node:
            raise ValueError(f"Value must have size {self.n_node}, got {value.size} instead.")
        self.uxds["node_elevation"][:] = value
        return

    @property
    def pix_mean(self) -> float:
        """
        The mean pixel size of the mesh.
        """
        if self._pix_mean is None and self.uxgrid is not None:
            self._compute_face_size()
        return self._pix_mean

    @property
    def pix_std(self) -> float:
        """
        The standard deviation of the pixel size of the mesh.
        """
        if self._pix_std is None and self.uxgrid is not None:
            self._compute_face_size()
        return self._pix_std

    @property
    def pix_min(self) -> float:
        """
        The minimum pixel size of the mesh.
        """
        if self._pix_min is None and self.uxgrid is not None:
            self._compute_face_size()
        return self._pix_min

    @property
    def pix_max(self) -> float:
        """
        The maximum pixel size of the mesh.
        """
        if self._pix_max is None and self.uxgrid is not None:
            self._compute_face_size()
        return self._pix_max

    @property
    def radius(self):
        """
        Radius of the target body.
        """
        return self.target.radius

    @property
    def area(self) -> float:
        """
        Total surface area of the target body.
        """
        if self._area is None:
            self._area = float(self.face_area.sum())
        return self._area

    @property
    def face_area(self) -> NDArray[np.float64]:
        """
        The areas of each face.

        Notes
        -----
        Unlike uxarray.Grid.face_area, this is in meters squared rather than normalized to a unit sphere.

        """
        if self._face_area is None:
            self._compute_face_size()
        return self._face_area

    @property
    def face_size(self) -> NDArray[np.float64]:
        """
        The effective size of each face in meters.

        This is simply the square root of the face area, but is useful for certain comparisons and is equivalent to the `pix` variable from CTEM
        """
        if self._face_size is None:
            self._compute_face_size()
        return self._face_size

    @property
    def smallest_length(self) -> float:
        """
        The smallest length of the mesh.
        """
        if self._smallest_length is None:
            self._smallest_length = float(np.min(self.face_size) * _SMALLFAC)
        return self._smallest_length

    @property
    def face_lat(self) -> NDArray[np.float64]:
        """
        Latitude of the center of each face in degrees.
        """
        return self.uxgrid.face_lat.values

    @property
    def face_lon(self) -> NDArray[np.float64]:
        """
        Longitude of the center of each face in degrees.
        """
        return self.uxgrid.face_lon.values

    @property
    def face_x(self) -> NDArray[np.float64]:
        """
        Cartesian x location of the center of each face in meters.
        """
        if self._face_x is None:
            self._face_x = self.uxgrid.face_x.values * self.radius
        return self._face_x

    @property
    def face_y(self) -> NDArray[np.float64]:
        """
        Cartesian y location of the center of each face in meters.
        """
        if self._face_y is None:
            self._face_y = self.uxgrid.face_y.values * self.radius
        return self._face_y

    @property
    def face_z(self) -> NDArray[np.float64]:
        """
        Cartesian z location of the center of each face in meters.
        """
        if self._face_z is None:
            self._face_z = self.uxgrid.face_z.values * self.radius
        return self._face_z

    @property
    def node_lat(self) -> NDArray[np.float64]:
        """
        Latitude of each node in degrees.
        """
        return self.uxgrid.node_lat.values

    @property
    def node_lon(self) -> NDArray[np.float64]:
        """
        Longitude of each node in degrees.
        """
        return self.uxgrid.node_lon.values

    @property
    def node_x(self) -> NDArray[np.float64]:
        """
        Cartesian x location of each node in meters.
        """
        if self._node_x is None:
            self._node_x = self.uxgrid.node_x.values * self.radius
        return self._node_x

    @property
    def node_y(self) -> NDArray[np.float64]:
        """
        Cartesian y location of each node in meters.
        """
        if self._node_y is None:
            self._node_y = self.uxgrid.node_y.values * self.radius
        return self._node_y

    @property
    def node_z(self) -> NDArray[np.float64]:
        """
        Cartesian z location of each node in meters.
        """
        if self._node_z is None:
            self._node_z = self.uxgrid.node_z.values * self.radius
        return self._node_z

    @property
    def face_indices(self) -> NDArray[np.int64]:
        """
        The indices of the faces of the surface.
        """
        return self.uxds.n_face.values

    @property
    def node_indices(self) -> NDArray[np.int64]:
        """
        The indices of the nodes of the surface.
        """
        return self.uxds.n_node.values

    @property
    def edge_indices(self) -> NDArray[np.int64]:
        """
        The indices of the edges of the surface.
        """
        if self._edge_indices is None:
            self._edge_indices = np.arange(self.uxgrid.n_edge, dtype=np.int64)
        return self._edge_indices

    def _compute_face_bins(self) -> None:
        """
        Compute the face bins based on the face areas. This is used to bin faces by their area for crater generation.
        """
        min_area = self.face_area.min()
        max_area = self.face_area.max()
        max_bin_index = np.ceil(np.log2(max_area / min_area)).astype(int)
        bins = [[] for _ in range(max_bin_index)]

        for face_index, area in enumerate(self.face_area):
            bin_index = np.floor(np.log2(area / min_area)).astype(int)
            bins[bin_index].append(face_index)

        self._face_bin_indices = [np.array(bins[i]) for i in range(max_bin_index) if len(bins[i]) > 0]

        self._face_bin_area = [np.sum(self.face_area[face_indices]) for face_indices in self.face_bin_indices]

        self._face_bin_argmin = [
            int(face_indices[np.argmin(self.face_area[face_indices])]) for face_indices in self._face_bin_indices
        ]

        self._face_bin_argmax = [
            int(face_indices[np.argmax(self.face_area[face_indices])]) for face_indices in self._face_bin_indices
        ]
        return

    @property
    def face_bin_indices(self) -> list[NDArray]:
        """
        Faces binned based on their area.

        All faces within a factor of 2 in area are in the same bin. This property returns a list of face indices lists for each bin.
        The keys are the bin indices, and the values are lists of face indices of faces within that bin.

        This is used when generating craters on surfaces with varying face sizes, so that the smallest crater is sized for the smallest face of a particular bin, rather than for the entire surface.
        """
        if self._face_bin_indices is None:
            self._compute_face_bins()

        return self._face_bin_indices

    @property
    def face_bin_area(self) -> list[float]:
        """
        The total area of all faces in each bin.
        """
        if self._face_bin_area is None:
            self._compute_face_bins()

        return self._face_bin_area

    @property
    def face_bin_argmin(self) -> list[int]:
        """
        The index of the smallest face in each bin.
        """
        if self._face_bin_argmin is None:
            self._compute_face_bins()

        return self._face_bin_argmin

    @property
    def face_bin_argmax(self) -> list[int]:
        """
        The index of the largest face in each bin.
        """
        if self._face_bin_argmax is None:
            self._compute_face_bins()

        return self._face_bin_argmax

    @property
    def face_bin_min_areas(self) -> list[float]:
        """
        The area of the smallest face in each bin.
        """
        if self._face_bin_argmin is None:
            self._compute_face_bins()

        return [float(self.face_area[face_index]) for face_index in self.face_bin_argmin]

    @property
    def face_bin_max_areas(self) -> list[float]:
        """
        The area of the largest face in each bin.
        """
        if self._face_bin_argmax is None:
            self._compute_face_bins()

        return [float(self.face_area[face_index]) for face_index in self.face_bin_argmax]

    @property
    def face_bin_min_sizes(self) -> list[float]:
        """
        The effective size of the smallest face in each bin.
        """
        if self._face_bin_argmin is None:
            self._compute_face_bins()

        return [float(self.face_size[face_index]) for face_index in self.face_bin_argmin]

    @property
    def face_bin_max_sizes(self) -> list[float]:
        """
        The effective size of the largest face in each bin.
        """
        if self._face_bin_argmax is None:
            self._compute_face_bins()

        return [float(self.face_size[face_index]) for face_index in self.face_bin_argmax]

    @property
    def n_face(self) -> int:
        """
        Total number of faces.
        """
        return int(self.uxgrid.n_face)

    @property
    def n_node(self) -> int:
        """
        Total number of nodes.
        """
        return int(self.uxgrid.n_node)

    @property
    def n_edge(self) -> int:
        """
        Total number of edges.
        """
        return int(self.uxgrid.n_edge)

    @property
    def n_nodes_per_face(self) -> NDArray[np.int64]:
        """
        The number of nodes that make up each face.

        Dimensions: `(n_node, )`
        """
        return self.uxgrid.n_nodes_per_face.values

    @property
    def n_max_face_faces(self) -> int:
        """
        The maximum number of faces that surround a face.
        """
        return int(self.uxgrid.n_max_face_faces)

    @property
    def edge_face_distance(self) -> NDArray[np.float64]:
        """
        Distances between the centers of the faces that saddle each edge in meters.
        """
        if self._edge_face_distance is None:
            self._edge_face_distance = surface_functions.compute_edge_distances(
                self.edge_face_connectivity, self.face_lon, self.face_lat, self.radius
            )
        return self._edge_face_distance

    @property
    def edge_length(self) -> NDArray[np.float64]:
        """
        Lengths of each edge in meters.

        Notes
        -----
        This is computed as the distance between the nodes that define the edge
        """
        if self._edge_lengths is None:
            self._edge_lengths = surface_functions.compute_edge_distances(
                self.edge_node_connectivity, self.node_lon, self.node_lat, self.radius
            )
        return self._edge_lengths

    @property
    def face_node_connectivity(self) -> NDArray[np.int64]:
        """
        Indices of the nodes that make up each face.

        Dimensions: `(n_face, n_max_face_nodes)`

        Nodes are in counter-clockwise order.
        """
        return self.uxgrid.face_node_connectivity.values

    @property
    def face_face_connectivity(self) -> NDArray[np.int64]:
        """
        Indices of the faces that surround each face.

        Dimensions: `(n_face, n_max_face_faces)`
        """
        return self.uxgrid.face_face_connectivity.values

    @property
    def face_edge_connectivity(self) -> NDArray[np.int64]:
        """
        Indices of the edges that surround each face.

        Dimensions (n_face, n_max_face_edges)

        """
        return self.uxgrid.face_edge_connectivity.values

    @property
    def node_face_connectivity(self) -> NDArray[np.int64]:
        """
        Indices of the faces that surround each node.

        Dimensions: `(n_node, n_max_node_faces)`
        """
        return self.uxgrid.node_face_connectivity.values

    @property
    def edge_face_connectivity(self) -> NDArray[np.int64]:
        """
        Indices of the faces that saddle each edge.

        Dimensions (n_edge, 2)

        """
        return self.uxgrid.edge_face_connectivity.values

    @property
    def edge_node_connectivity(self) -> NDArray[np.int64]:
        """
        Indices of the nodes that define each edge.

        Dimensions (n_edge, 2)

        """
        return self.uxgrid.edge_node_connectivity.values

    @property
    def face_tree(self):
        if self._face_tree is None:
            self._face_tree = self.uxgrid.get_ball_tree(
                "face centers",
                distance_metric="haversine",
                coordinate_system="spherical",
                reconstruct=True,
            )

        return self._face_tree

    @property
    def node_tree(self):
        if self._node_tree is None:
            self._node_tree = self.uxgrid.get_ball_tree(
                "nodes",
                distance_metric="haversine",
                coordinate_system="spherical",
                reconstruct=True,
            )

        return self._node_tree

    @property
    def edge_tree(self):
        if self._edge_tree is None:
            self._edge_tree = self.uxgrid.get_ball_tree(
                "edge centers",
                distance_metric="haversine",
                coordinate_system="spherical",
                reconstruct=True,
            )

        return self._edge_tree


class LocalSurface:
    """
    Generates a regional view of a subset of the surface mesh without making copies of any of the data.

    Parameters
    ----------
    surface : Surface
        The surface object that contains the mesh data.
    face_indices : NDArray | slice
        The indices of the faces to include in the view.
    node_indices : NDArray | slice | None, optional
        The indices of the nodes to include in the view. If None, all nodes connected to the faces will be extracted when required
    edge_indices : NDArray | slice | None, optional
        The indices of the edges to include in the view. If None, all edges connected to the faces will be extracted when required.
    location : tuple[float, float] | None, optional
        The location of the center of the view in degrees. This is intended to be passed via the extract_region method of Surface.
    region_radius : FloatLike | None, optional
        The radius of the region to include in the view in meters. This is intended to be passed via the extract_region method of Surface.
    """

    def __init__(
        self,
        surface: Surface,
        face_indices: NDArray | slice,
        node_indices: NDArray | slice | None = None,
        edge_indices: NDArray | slice | None = None,
        location: tuple[float, float] | None = None,
        region_radius: FloatLike | None = None,
        **kwargs: Any,
    ):
        object.__setattr__(self, "_surface", Surface.maker(surface))
        object.__setattr__(self, "_face_indices", face_indices)
        object.__setattr__(self, "_node_indices", node_indices)
        object.__setattr__(self, "_edge_indices", edge_indices)
        object.__setattr__(self, "_region_radius", region_radius)
        object.__setattr__(self, "_location", None)
        object.__setattr__(self, "_area", None)
        object.__setattr__(self, "_n_edge", None)
        object.__setattr__(self, "_n_face", None)
        object.__setattr__(self, "_n_node", None)
        object.__setattr__(self, "_face_distance", None)
        object.__setattr__(self, "_node_distance", None)
        object.__setattr__(self, "_face_bearing", None)
        object.__setattr__(self, "_node_bearing", None)
        object.__setattr__(self, "_edge_face_connectivity", None)
        object.__setattr__(self, "_face_edge_connectivity", None)
        object.__setattr__(self, "_face_node_connectivity", None)
        object.__setattr__(self, "_face_face_connectivity", None)
        object.__setattr__(self, "_node_face_connectivity", None)

        if location is not None:
            self._location = validate_and_normalize_location(location)
            self._face_distance, self._node_distance = self.calculate_face_and_node_distances()
            self._face_bearing, self._node_bearing = self.calculate_face_and_node_bearings()
        return

    def __str__(self) -> str:
        """
        String representation of the LocalSurface object.
        """
        base = "<LocalSurface>"
        if self.location:
            base += f"\nLocation: {self.location[0]:.2f}°, {self.location[1]:.2f}°"

        if self.region_radius:
            base += f"\nRegion Radius: {format_large_units(self.region_radius, quantity='length')}"

        return f"{base}\nNumber of faces: {self.n_face}\nNumber of nodes: {self.n_node}"

    def add_data(
        self,
        name: str,
        data: FloatLike | NDArray,
        long_name: str | None = None,
        units: str | None = None,
        isfacedata: bool = True,
        overwrite: bool = False,
    ) -> None:
        """
        Adds new data to the surface.

        Parameters
        ----------
        name : str
            Name of the data variable. This will also be used as the data file name.
        data : scalar or array-like
            Data file to be saved. If data is a scalar, then the data file will be filled with that value. If data is an array, then the data file will be filled with the array values. The data array must have the same size as the number of faces or nodes in the grid.
        long_name : str, optional
            Long name of the data variable that will be saved as an attribute if this is new data. If the data already exists on the surface, this will be ignored.
        units : str, optional
            Units of the data variable that will be saved as an attribute if this is new data. If the data already exists on the surface, this will be ignored.
        isfacedata : bool, optional, default True
            Flag to indicate whether the data is face data or node data. This is only needed if `data` is a scalar, otherwise it is ignored
        overwrite : bool, optional, default False
            By default, new data is added to the old data. This flag indicates that the data should be overwritten, replacing any old data with the new data.

        Returns
        -------
        None
        """
        # Check if the data is a scalar or an array
        if np.isscalar(data):
            n = self.n_face if isfacedata else self.n_node
            data = np.full(n, data)
        elif isinstance(data, list):
            data = np.array(data)
        else:
            data = np.asarray(data)
        if data.size == self.n_face:
            isfacedata = True
            indices = self.face_indices
        elif data.size == self.n_node:
            isfacedata = False
            indices = self.node_indices
        else:
            raise ValueError("data must be a scalar or an array with the same size as the number of faces or nodes in the grid")

        if name not in self.surface.uxds.data_vars:
            self.surface._add_new_data(name, data=0.0, long_name=long_name, units=units, isfacedata=isfacedata)

        # This prevents concurrent writes to the same data variable when used in threading
        with surface_lock:
            if overwrite:
                self.surface.uxds[name].data[indices] = data
            else:
                self.surface.uxds[name].data[indices] += data

        return

    def update_elevation(
        self,
        new_elevation: ArrayLike | FloatLike,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Update the elevation data for the target's surface mesh. This method will determine whether to update the node or face data (or both) depending on the size of the input data. If a scalar is passed, both node and face elevation will be updated to that value. By default, the elevation data will be added to the existing data. If `overwrite` is set to True, the existing data will be replaced with the new data.

        Parameters
        ----------
        new_elevation : ArrayLike | FloatLike
            Elevation to be added (or replaced, if overwrite is True). This can be a scalar, an array with the same size as the number of faces, an array with the same size as the number of nodes, or an array with the same size as the number of faces + nodes.
        overwrite : bool, optional
            If True, the existing data will be replaced with the new data. Default is False.
        **kwargs : Any
            Additional keyword arguments.

        Notes
        -----
        When passing combined data, the first part of the array will be used for face elevation and the second part for node elevation.
        """

        def raise_invalid_elevation_error():
            raise ValueError(
                "new_elev must be None, a scalar, or an array with the same size as the number of nodes, faces, or nodes+faces"
            )

        try:
            new_elevation = np.asarray(new_elevation)

            if np.asarray(new_elevation).size == 1:
                update_node = True
                update_face = True
                new_face_elev = np.full(self.n_face, new_elevation.item())
                new_node_elev = np.full(self.n_node, new_elevation.item())
            else:
                if new_elevation.size == self.n_face:
                    update_face = True
                    update_node = False
                    new_face_elev = new_elevation
                elif new_elevation.size == self.n_node:
                    update_face = False
                    update_node = True
                    new_node_elev = new_elevation
                elif new_elevation.size == self.n_face + self.n_node:
                    update_face = True
                    update_node = True
                    new_face_elev = new_elevation[: self.n_face]
                    new_node_elev = new_elevation[self.n_face :]
                else:
                    raise_invalid_elevation_error()
        except Exception as e:
            raise ValueError("new_elev must be None, a scalar, or an array") from e

        if update_face:
            self.add_data(name="face_elevation", data=new_face_elev, overwrite=overwrite)
        if update_node:
            self.add_data(name="node_elevation", data=new_node_elev, overwrite=overwrite)

        return

    def apply_diffusion(self, kdiff: FloatLike | NDArray) -> None:
        """
        Apply diffusion to the surface.

        Parameters
        ----------
        kdiff : float or array-like
            The degradation state of the surface, which is the product of diffusivity and time. It can be a scalar or an array of the same size as the number of faces in the grid.
            If it is a scalar, the same value is applied to all faces. If it is an array, it must have the same size as the number of faces in the grid.
            The value of kdiff must be greater than 0.0.

        """
        if np.isscalar(kdiff):
            kdiff = np.full(self.n_face, kdiff)
        elif kdiff.size != self.n_face:
            raise ValueError("kdiff must be a scalar or an array with the same size as the number of faces in the grid")
        if np.any(kdiff < 0.0):
            raise ValueError("kdiff must be greater than 0.0")
        kdiff = np.asarray(kdiff, dtype=np.float64)
        kdiffmax = np.max(kdiff)

        if abs(kdiffmax) < _VSMALL:
            return

        delta_face_elevation = surface_functions.apply_diffusion(
            face_kappa=kdiff,
            face_elevation=self.face_elevation,
            face_area=self.face_area,
            edge_face_connectivity=self.edge_face_connectivity,
            edge_face_distance=self.edge_face_distance,
            edge_length=self.edge_length,
        )
        self.update_elevation(delta_face_elevation)
        self.add_data("ejecta_thickness", delta_face_elevation)
        self.interpolate_node_elevation_from_faces()
        return

    def slope_collapse(self, critical_slope_angle: FloatLike = 35.0) -> NDArray:
        """
        Collapse all slopes larger than the critical slope angle.

        Parameters
        ----------
        critical_slope_angle : float
            The critical slope angle (angle of repose) in degrees.
        """
        try:
            critical_slope = np.tan(np.deg2rad(critical_slope_angle))
        except ValueError as e:
            raise ValueError("critical_slope_angle must be between 0 and 90 degrees") from e

        delta_face_elevation = surface_functions.slope_collapse(
            critical_slope=critical_slope,
            face_elevation=self.face_elevation,
            face_area=self.face_area,
            edge_face_connectivity=self.edge_face_connectivity,
            face_edge_connectivity=self.face_edge_connectivity,
            edge_face_distance=self.edge_face_distance,
            edge_length=self.edge_length,
        )
        self.update_elevation(delta_face_elevation)
        self.add_data("ejecta_thickness", delta_face_elevation)
        self.interpolate_node_elevation_from_faces()

    def compute_slope(self) -> NDArray[np.float64]:
        """
        Compute the slope of the surface.

        Returns
        -------
        NDArray[np.float64]
            The slope of all faces in degrees.
        """
        slope = surface_functions.compute_slope(
            face_elevation=self.face_elevation,
            edge_face_connectivity=self.edge_face_connectivity,
            face_edge_connectivity=self.face_edge_connectivity,
            edge_face_distance=self.edge_face_distance,
            edge_length=self.edge_length,
        )

        return np.rad2deg(np.arctan(slope))

    def apply_noise(
        self,
        model: str = "turbulence",
        noise_width: FloatLike = 1000e3,
        noise_height: FloatLike = 1e3,
        **kwargs: Any,
    ) -> None:
        """
        Apply noise to the node elevations of the surface view.

        Parameters
        ----------
        noise_width : float
            The spatial wavelength of the noise.
        noise_height : float
            The amplitude of the noise.
        """
        num_octaves = kwargs.pop("num_octaves", 12)
        anchor = kwargs.pop(
            "anchor",
            self.surface.rng.uniform(0, 2 * np.pi, size=(num_octaves, 3)),
        )
        x = np.concatenate([self.face_x, self.node_x]) / self.surface.radius
        y = np.concatenate([self.face_y, self.node_y]) / self.surface.radius
        z = np.concatenate([self.face_z, self.node_z]) / self.surface.radius

        if model == "turbulence":
            noise = surface_functions.turbulence_noise(
                x=x,
                y=y,
                z=z,
                noise_height=noise_height / self.surface.radius,
                noise_width=noise_width / self.surface.radius,
                freq=2.0,
                pers=0.5,
                anchor=anchor,
                seed=self.surface.rng.integers(0, 2**32 - 1),
            )
        else:
            raise ValueError(f"Unknown noise model: {model}")
        # Compute the weighted mean to ensure volume conservation
        mean = np.sum(noise[: self.n_face] * self.face_area) / self.area
        noise -= mean
        self.update_elevation(noise)
        return

    def calculate_face_and_node_distances(
        self, location: tuple[float, float] | None = None
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Computes the distances between nodes and faces and a given location.

        Parameters
        ----------
        location : tuple[float, float], option
            tuple containing the longitude and latitude of the location in degrees. If None, the location of the view center is used if it is set.

        Returns
        -------
        NDArray
            Array of face distances in meters.
        NDArray
            Array of node distances in meters.
        """
        if location is None:
            if self.location is None:
                raise ValueError("location must be set.")
            location = self.location

        if len(location) == 1:
            location = location.item()
        if len(location) != 2:
            raise ValueError("location must be a single pair of (longitude, latitude).")
        location = validate_and_normalize_location(location)
        lon1 = np.deg2rad(location[0])
        lat1 = np.deg2rad(location[1])
        node_lon2 = np.deg2rad(self.node_lon)
        node_lat2 = np.deg2rad(self.node_lat)
        face_lon2 = np.deg2rad(self.face_lon)
        face_lat2 = np.deg2rad(self.face_lat)
        return self._calculate_distance(lon1, lat1, face_lon2, face_lat2), self._calculate_distance(
            lon1, lat1, node_lon2, node_lat2
        )

    def calculate_face_and_node_bearings(self, location: tuple[float, float] | None = None) -> tuple[NDArray, NDArray]:
        """
        Computes the initial bearing between nodes and faces and a given location.

        Parameters
        ----------
        location : tuple[float, float]
            tuple containing the longitude and latitude of the location in degrees.

        Returns
        -------
        NDArray
            Array of initial bearings for each face in radians.
        NDArray
            Array of initial bearings for each node in radians.
        """
        if location is None:
            if self.location is None:
                raise ValueError("location must be set.")
            location = self.location

        if len(location) == 1:
            location = location.item()
        if len(location) != 2:
            raise ValueError("location must be a single pair of (longitude, latitude).")
        location = validate_and_normalize_location(location)
        lon1 = np.deg2rad(location[0])
        lat1 = np.deg2rad(location[1])
        node_lon2 = np.deg2rad(self.node_lon)
        node_lat2 = np.deg2rad(self.node_lat)
        face_lon2 = np.deg2rad(self.face_lon)
        face_lat2 = np.deg2rad(self.face_lat)
        return (
            surface_functions.calculate_bearing(lon1, lat1, face_lon2, face_lat2),
            surface_functions.calculate_bearing(lon1, lat1, node_lon2, node_lat2),
        )

    def interpolate_node_elevation_from_faces(self) -> None:
        """
        Update node elevations by area-weighted averaging of adjacent face elevations.

        For each node, the elevation is computed as the area-weighted average of the elevations
        of the surrounding faces.

        Returns
        -------
        None
        """
        node_elevation = surface_functions.interpolate_node_elevation_from_faces(
            face_area=self.face_area,
            face_elevation=self.face_elevation,
            node_face_connectivity=self.node_face_connectivity,
        )
        self.update_elevation(node_elevation, overwrite=True)
        return

    def get_reference_surface(self, reference_radius: float, **kwargs: Any) -> NDArray[np.float64]:
        """
        Calculate the orientation of a hemispherical cap that represents the average surface within a given region.

        Parameters
        ----------
        reference_radius : float
            The radius of the reference region to compute the average over in meters.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        NDArray[np.float64]
            The face and node elevation points of the reference sphere, or the original elevation points is th reference region is too small
        """

        def _find_reference_elevations(region_coords, region_elevation):
            # Perform the curve fitting to get the best fitting spherical cap for the reference surface

            region_surf = self.surface._compute_elevation_to_cartesian(region_coords, region_elevation)

            # Initial guess for the sphere center and radius
            guess_radius = 1.0 + region_elevation.mean()
            initial_guess = [0, 0, 0, guess_radius]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", OptimizeWarning)
                try:
                    bounds = ([-1.0, -1.0, -1.0, 0.5], [1.0, 1.0, 1.0, 2.0])
                    fit_result, _ = curve_fit(
                        self.surface._sphere_function,
                        region_surf,
                        np.zeros_like(region_elevation),
                        p0=initial_guess,
                        bounds=bounds,
                    )
                except Exception:
                    fit_result = initial_guess

            reference_sphere_center = fit_result[:3]
            reference_sphere_radius = fit_result[3]

            # Find the point along the original vector that intersects the sphere
            f_vec = region_coords / self.surface.radius
            a = f_vec[:, 0] ** 2 + f_vec[:, 1] ** 2 + f_vec[:, 2] ** 2
            b = -2 * (
                f_vec[:, 0] * reference_sphere_center[0]
                + f_vec[:, 1] * reference_sphere_center[1]
                + f_vec[:, 2] * reference_sphere_center[2]
            )
            c = np.dot(reference_sphere_center, reference_sphere_center) - reference_sphere_radius**2
            sqrt_term = b**2 - 4 * a * c
            valid = ~np.isnan(a) & (sqrt_term >= 0.0)

            # Initialize t with default value
            t = np.full_like(a, 1.0)

            # Calculate square root only for valid terms
            sqrt_valid_term = np.sqrt(np.where(valid, sqrt_term, 0))

            # Apply the formula only where valid
            t = np.where(valid, (-b + sqrt_valid_term) / (2 * a), t)
            if np.any(t[valid] < 0):
                t = np.where(valid & (t < 0), (-b - sqrt_valid_term) / (2 * a), t)

            elevations = self.surface.radius * (t * np.linalg.norm(f_vec, axis=1) - 1)
            return elevations

        # Find cells within the crater radius
        faces_within_region = self.face_distance <= reference_radius
        nodes_within_region = self.node_distance <= reference_radius
        points_within_region = np.concatenate([faces_within_region, nodes_within_region])

        elevation = np.concatenate([self.face_elevation, self.node_elevation])

        n_reference = np.sum(faces_within_region) + np.sum(nodes_within_region)

        # if there are not enough faces or nodes within the radius, return the original elevations and use that as the reference
        if n_reference < 5 or not nodes_within_region.any():
            return elevation

        combo_grid = np.column_stack(
            (
                np.concatenate([self.face_x, self.node_x]),
                np.concatenate([self.face_y, self.node_y]),
                np.concatenate([self.face_z, self.node_z]),
            )
        )
        # find_refence_elevations expects coordinates to be on the unit sphere, so we need to normalize
        region_coords = combo_grid[points_within_region] / self.surface.radius
        region_elevation = elevation[points_within_region] / self.surface.radius

        reference_elevation = elevation
        reference_elevation[points_within_region] = _find_reference_elevations(region_coords, region_elevation)

        return reference_elevation

    def extract_subregion(self, subregion_radius: FloatLike):
        """
        Extract a subset of the LocalSurface region with a smaller radius than the original region.

        Parameters
        ----------
        subregion_radius : float
            The radius of the subregion to extract in meters.

        Returns
        -------
        LocalSurface
            A LocalSurface object containing a view of the regional grid.

        """
        if subregion_radius > self.region_radius:
            raise ValueError("subregion_radius must be smaller than the original region radius")

        if subregion_radius < self.region_radius:
            if isinstance(self.face_indices, slice):
                face_indices = np.arange(self.n_face)[self.face_indices][self.face_distance <= subregion_radius]
            else:
                face_indices = self.face_indices[self.face_distance <= subregion_radius]
            if face_indices.size == 0:
                return None

            # First select edges and nodes that are attached to these faces
            edge_indices = np.unique(self.surface.face_edge_connectivity[face_indices].ravel())
            edge_indices = edge_indices[edge_indices != INT_FILL_VALUE]

            node_indices = np.unique(self.surface.face_node_connectivity[face_indices].ravel())
            node_indices = node_indices[node_indices != INT_FILL_VALUE]

            # Now add in all faces that are connected to anything inside the region, so that the outermost border of the local region has a buffer of faces
            # These are needed for diffusion calculations
            neighbor_faces = self.surface.face_face_connectivity[face_indices]
            neighbor_faces = neighbor_faces[neighbor_faces != INT_FILL_VALUE]
            node_faces = self.surface.node_face_connectivity[node_indices]
            node_faces = node_faces[node_faces != INT_FILL_VALUE]
            edge_faces = self.surface.edge_face_connectivity[edge_indices]
            edge_faces = edge_faces[edge_faces != INT_FILL_VALUE]
            face_indices = np.unique(np.concatenate((face_indices, neighbor_faces, node_faces, edge_faces)))
        else:
            face_indices = self.face_indices
            edge_indices = self.edge_indices
            node_indices = self.node_indices

        return LocalSurface(
            surface=self.surface,
            face_indices=face_indices,
            node_indices=node_indices,
            edge_indices=edge_indices,
            location=self.location,
            region_radius=subregion_radius,
        )

    def compute_volume(self, elevation: NDArray) -> NDArray:
        """
        Compute the volume of an array of elevation points.

        Parameters
        ----------
        elevation : NDArray
            The elevation values to compute. This should be the same size as the number of faces in the grid.

        Returns
        -------
        float
            The volume of the elevation points
        """
        if elevation.size != self.n_face:
            raise ValueError("elevation must be an array with the same size as the number of faces in the grid")
        return np.sum(elevation * self.face_area)

    def elevation_to_cartesian(self, element="face") -> NDArray[np.float64]:
        """
        Convert either the face or node elevations to Cartesian coordinates.

        Parameters
        ----------
        element : str, optional
            The type of element to convert. Can be "face" or "node". Default is "face".

        Returns
        -------
        NDArray[np.float64, 3]
            The Cartesian coordinates of the face elevations.
        """
        if element not in ("face", "node"):
            raise ValueError("element must be either 'face' or 'node'")
        if element == "face":
            position = np.column_stack((self.face_x, self.face_y, self.face_z))
            elevation = self.face_elevation
        elif element == "node":
            position = np.column_stack((self.node_x, self.node_y, self.node_z))
            elevation = self.node_elevation

        return self.surface._compute_elevation_to_cartesian(position, elevation)

    def _calculate_distance(
        self,
        lon1: FloatLike,
        lat1: FloatLike,
        lon2: FloatLike | ArrayLike,
        lat2: FloatLike | ArrayLike,
    ) -> NDArray[np.float64]:
        """
        Calculate the great circle distance between one point and one or more other points in meters.

        Parameters
        ----------
        lon1 : FloatLike
            Longitude of the first point in radians.
        lat1 : FloatLike
            Latitude of the first point in radians.
        lon2 : FloatLike or ArrayLike
            Longitude of the second point or array of points in radians.
        lat2 : FloatLike or ArrayLike
            Latitude of the second point or array of points in radians.

        Returns
        -------
        NDArray
            Great circle distance between the two points in meters.

        Notes
        -----
        This is a wrapper for a compiled Rust function and is intended to be used as a helper to calculate_face_and_node_distances.
        """
        return self.surface._calculate_distance(lon1, lat1, lon2, lat2)

    @staticmethod
    def _calculate_bearing(
        lon1: FloatLike,
        lat1: FloatLike,
        lon2: FloatLike | ArrayLike,
        lat2: FloatLike | ArrayLike,
    ) -> NDArray[np.float64]:
        """
        Calculate the initial bearing from one point to one or more other points in radians.

        Parameters
        ----------
        lon1 : FloatLike
            Longitude of the first point in radians.
        lat1 : FloatLike
            Latitude of the first point in radians.
        lon2 : FloatLike or ArrayLike
            Longitude of the second point or array of points in radians.
        lat2 : FloatLike or ArrayLike
            Latitude of the second point or array of points in radians.

        Returns
        -------
        NDArray
            Initial bearing from the first point to the second point or points in radians.

        Notes
        -----
        This is intended to be used as a helper to calculate_face_and_node_bearings.
        """
        # Calculate differences in coordinates
        dlon = np.mod(lon2 - lon1 + np.pi, 2 * np.pi) - np.pi

        # Haversine formula calculations
        x = np.sin(dlon) * np.cos(lat2)
        y = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(dlon)
        initial_bearing = np.arctan2(x, y)

        # Normalize bearing to 0 to 2*pi
        initial_bearing = (initial_bearing + 2 * np.pi) % (2 * np.pi)

        return initial_bearing

    @staticmethod
    def _remap_connectivity_to_local(
        connectivity_array: np.ndarray,
        row_indices: np.ndarray,
        value_indices: np.ndarray,
        total_value_size: int,
        fill_value: int = INT_FILL_VALUE,
    ) -> np.ndarray:
        """
        Remap a 2D connectivity array where the rows are indexed by one entity (e.g., edges) and the values are global references to another entity (e.g., faces).

        Only rows in `row_indices` are retained, and their values are remapped from global to local indices using `value_indices`.

        Parameters
        ----------
        connectivity_array : np.ndarray
            The global 2D connectivity array of shape (N, K).
        row_indices : np.ndarray
            Global indices indicating which rows to retain.
        value_indices : np.ndarray
            Global indices of values to remap (e.g., face indices).
        total_value_size : int
            Total number of global items in the value domain (e.g., total number of faces).
        fill_value : int, optional
            Fill value indicating invalid entries (default is INT_FILL_VALUE).

        Returns
        -------
        np.ndarray
            A new 2D connectivity array of shape (len(row_indices), K) with local indices or fill_value where not applicable.
        """
        if isinstance(row_indices, slice) and row_indices == slice(None):  # This is a global slice. Just return the original array
            return connectivity_array
        if isinstance(value_indices, slice):
            value_indices = np.arange(total_value_size)[value_indices]

        # Build global-to-local index mapping
        mapping = np.full(total_value_size, fill_value, dtype=np.int64)
        mapping[value_indices] = np.arange(value_indices.size)

        subset = connectivity_array[row_indices]
        remapped = np.full_like(subset, fill_value)
        valid = subset != fill_value
        remapped[valid] = mapping[subset[valid]]

        return remapped

    @property
    def surface(self) -> Surface:
        """
        The surface object that contains the mesh data.
        """
        return self._surface

    @property
    def n_edge(self) -> int:
        """
        The number of edges in the view.
        """
        if self._n_edge is None:
            if isinstance(self._edge_indices, slice):
                self._n_edge = int(self._surface._uxds.uxgrid.n_edge[self._edge_indices].size)
            else:
                self._n_edge = self._edge_indices.size
        return self._n_edge

    @property
    def n_face(self) -> int:
        """
        The number of faces in the view.
        """
        if self._n_face is None:
            if isinstance(self.face_indices, slice):
                self._n_face = int(self.surface.face_elevation[self.face_indices].size)
            elif isinstance(self.face_indices, np.ndarray):
                self._n_face = int(self.face_indices.size)

        return self._n_face

    @property
    def n_node(self) -> int:
        """
        The number of nodes in the view.
        """
        if self._n_node is None:
            if isinstance(self.node_indices, slice):
                self._n_node = int(self.surface.node_elevation[self.node_indices].size)
            elif isinstance(self.node_indices, np.ndarray):
                self._n_node = int(self.node_indices.size)

        return self._n_node

    @property
    def face_elevation(self) -> NDArray:
        """
        The elevation of the faces.
        """
        return self.surface.face_elevation[self.face_indices]

    @face_elevation.setter
    def face_elevation(self, value: NDArray) -> None:
        """
        Set the elevation of the faces.

        Parameters
        ----------
        value : NDArray
            The elevation values to set for the faces.
        """
        if value.size != self.n_face:
            raise ValueError(f"Value must have size {self.n_face}, got {value.size} instead.")
        self.surface.face_elevation[self.face_indices] = value
        return

    @property
    def node_elevation(self) -> NDArray:
        """
        The elevation of the nodes.
        """
        return self.surface.node_elevation[self.node_indices]

    @node_elevation.setter
    def node_elevation(self, value: NDArray) -> None:
        """
        Set the elevation of the nodes.

        Parameters
        ----------
        value : NDArray
            The elevation values to set for the nodes.
        """
        if value.size != self.n_node:
            raise ValueError(f"Value must have size {self.n_node}, got {value.size} instead.")
        self.surface.node_elevation[self.node_indices] = value
        return

    @property
    def location(self) -> tuple[float, float]:
        """
        The location of the center of the view.
        """
        return self._location

    @property
    def region_radius(self) -> FloatLike:
        """
        The radius of the region to include in the view in meters.
        """
        return self._region_radius

    @property
    def face_bearing(self) -> NDArray:
        """
        The initial bearing from the location to the faces.
        """
        return self._face_bearing

    @property
    def node_bearing(self) -> NDArray:
        """
        The initial bearing from the location to the nodes.
        """
        return self._node_bearing

    @property
    def face_distance(self) -> NDArray:
        """
        The distance from the location to the faces.
        """
        return self._face_distance

    @property
    def node_distance(self) -> NDArray:
        """
        The distance from the location to the nodes.
        """
        return self._node_distance

    @property
    def face_area(self) -> NDArray:
        """
        The areas of the faces.
        """
        return self.surface.face_area[self.face_indices]

    @property
    def face_lat(self) -> NDArray:
        """
        Latitude of the center of the faces in degrees.
        """
        return self.surface.face_lat[self.face_indices]

    @property
    def face_lon(self) -> NDArray:
        """
        Longitude of the center of the faces in degrees.
        """
        return self.surface.face_lon[self.face_indices]

    @property
    def face_x(self) -> NDArray:
        """
        Cartesian x location of the center of the faces in meters.
        """
        return self.surface.face_x[self.face_indices]

    @property
    def face_y(self) -> NDArray:
        """
        Cartesian y location of the center of the faces in meters.
        """
        return self.surface.face_y[self.face_indices]

    @property
    def face_z(self) -> NDArray:
        """
        Cartesian z location of the center of the faces in meters.
        """
        return self.surface.face_z[self.face_indices]

    @property
    def edge_face_distance(self) -> NDArray:
        """
        Distances between the edges and the faces.

        Dimensions: `(n_edge)`
        """
        return self.surface.edge_face_distance[self.edge_indices]

    @property
    def edge_length(self) -> NDArray:
        """
        Lengths of the edges in meters.

        Dimensions: `(n_edge)`
        """
        return self.surface.edge_length[self.edge_indices]

    @property
    def node_lat(self) -> NDArray:
        """
        Latitude of the nodes in degrees.
        """
        return self.surface.node_lat[self.node_indices]

    @property
    def node_lon(self) -> NDArray:
        """
        Longitude of the nodes in degrees.
        """
        return self.surface.node_lon[self.node_indices]

    @property
    def node_x(self) -> NDArray:
        """
        Cartesian x location of the nodes in meters.
        """
        return self.surface.node_x[self.node_indices]

    @property
    def node_y(self) -> NDArray:
        """
        Cartesian y location of the nodes in meters.
        """
        return self.surface.node_y[self.node_indices]

    @property
    def node_z(self) -> NDArray:
        """
        Cartesian z location of the nodes in meters.
        """
        return self.surface.node_z[self.node_indices]

    @property
    def area(self) -> float:
        """
        The total area of the faces in the view.
        """
        if self._area is None:
            self._area = float(self.face_area.sum())
        return self._area

    @property
    def face_indices(self) -> NDArray:
        """
        The indices of the faces in the view.
        """
        if self._face_indices is None:
            raise ValueError("face_indices must be set to use this object.")
        return self._face_indices

    @property
    def node_indices(self) -> NDArray:
        """
        The indices of the nodes in the view.
        """
        if self._node_indices is None:
            self._node_indices = np.unique(self.surface.face_node_connectivity[self.face_indices].ravel())
            self._node_indices = self._node_indices[self._node_indices != INT_FILL_VALUE]

        return self._node_indices

    @property
    def edge_indices(self) -> NDArray:
        """
        The indices of the edges in the view.
        """
        if self._edge_indices is None:
            self._edge_indices = np.unique(self.surface.face_edge_connectivity[self.face_indices].ravel())
            self._edge_indices = self._edge_indices[self._edge_indices != INT_FILL_VALUE]
        return self._edge_indices

    @property
    def edge_face_connectivity(self) -> NDArray:
        """
        Local indices of the faces that make up the edges.

        Dimensions: `(n_edge, 2)`
        """
        if self._edge_face_connectivity is None:
            self._edge_face_connectivity = self._remap_connectivity_to_local(
                connectivity_array=self.surface.edge_face_connectivity,
                row_indices=self.edge_indices,
                value_indices=self.face_indices,
                total_value_size=self.surface.n_face,
            )
        return self._edge_face_connectivity

    @property
    def face_edge_connectivity(self) -> NDArray:
        """
        Local indices of the edges that make up the faces.

        Dimensions: `(n_face, n_max_face_edges)`
        """
        if self._face_edge_connectivity is None:
            self._face_edge_connectivity = self._remap_connectivity_to_local(
                connectivity_array=self.surface.face_edge_connectivity,
                row_indices=self.face_indices,
                value_indices=self.edge_indices,
                total_value_size=self.surface.n_edge,
            )
        return self._face_edge_connectivity

    @property
    def face_node_connectivity(self) -> NDArray:
        """
        Local indices of the nodes that make up the faces.

        Dimensions: `(n_face, n_max_face_nodes)`

        Nodes are in counter-clockwise order.
        """
        if self._face_node_connectivity is None:
            self._face_node_connectivity = self._remap_connectivity_to_local(
                connectivity_array=self.surface.face_node_connectivity,
                row_indices=self.face_indices,
                value_indices=self.node_indices,
                total_value_size=self.surface.n_node,
            )
        return self._face_node_connectivity

    @property
    def face_face_connectivity(self) -> NDArray:
        """
        Indices of the faces that surround the faces.

        Dimensions: `(n_face, n_max_face_faces)`
        """
        if self._face_face_connectivity is None:
            self._face_face_connectivity = self._remap_connectivity_to_local(
                connectivity_array=self.surface.face_face_connectivity,
                row_indices=self.face_indices,
                value_indices=self.face_indices,
                total_value_size=self.surface.n_face,
            )
        return self._face_face_connectivity

    @property
    def node_face_connectivity(self) -> NDArray:
        """
        Indices of the faces that surround the nodes.

        Dimensions: `(n_node, n_max_node_faces)`
        """
        if self._node_face_connectivity is None:
            self._node_face_connectivity = self._remap_connectivity_to_local(
                connectivity_array=self.surface.node_face_connectivity,
                row_indices=self.node_indices,
                value_indices=self.face_indices,
                total_value_size=self.surface.n_face,
            )
        return self._node_face_connectivity

    @property
    def edge_node_connectivity(self) -> NDArray:
        """
        Local indices of the nodes that make up the edges.

        Dimensions: `(n_edge, 2)`
        """
        if self._edge_node_connectivity is None:
            self._edge_node_connectivity = self._remap_connectivity_to_local(
                connectivity_array=self.surface.edge_node_connectivity,
                row_indices=self.edge_indices,
                value_indices=self.node_indices,
                total_value_size=self.surface.n_node,
            )
        return self._edge_node_connectivity


import_components(__name__, __path__, ignore_private=True)



"""
Copyright 2025 - David Minton
This file is part of Cratermaker.
Cratermaker is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Cratermaker is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Cratermaker.
If not, see: https://www.gnu.org/licenses.
"""

import warnings

from ._version import version as __version__
from .components.morphology import Morphology
from .components.production import Production
from .components.projectile import Projectile
from .components.scaling import Scaling
from .components.surface import Surface
from .components.target import Target
from .core.crater import Crater
from .core.simulation import Simulation

__all__ = [
    "Crater",
    "Morphology",
    "Production",
    "Projectile",
    "Scaling",
    "Surface",
    "Target",
    "Simulation",
    "__version__",
]


# This will suppress the warning issued by xarray starting in version 2023.12.0 about the change in the API regarding .dims
# The API change does not affect the functionality of the code, so we can safely ignore the warning
warnings.filterwarnings("ignore", category=FutureWarning, module="xarray")
warnings.filterwarnings("ignore", category=FutureWarning, module="uxarray")
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, module="geopandas._compat"
)
warnings.filterwarnings(
    "ignore", category=RuntimeWarning, message="numpy.ndarray size changed"
)
warnings.filterwarnings(
    "ignore", category=UserWarning, module="uxarray.grid.coordinates"
)
warnings.filterwarnings("ignore", category=DeprecationWarning, module="tqdm.std")



"""
Sample a power law and lunar Neukum Production Function
========================================================

.. rubric:: By Austin Blevins and David Minton

In this example, we use the :func:`Production.sample` method to sample populations of craters from a production function. We will sample 1000 craters from each production function and repeat this 100 times. The population samples are then plotted against the expected value given by :func:`Production.function`
"""

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np

from cratermaker import Production, Target

fig, axs = plt.subplots(1, 2, figsize=(8, 7))
axes = dict(zip(["Power Law", "NPF (Moon)"], axs))

production = {
    "Power Law": Production.maker("powerlaw"),
    "NPF (Moon)": Production.maker("neukum", version="Moon"),
}

target = Target.maker("Moon")
area = 4 * np.pi * target.radius**2
age = 4100.0
x_min = 1e0
x_max = 1e5
y_min = 1e0
y_max = 1e4
diameter_range = (2e3, 10000e3)  # Range of diameters to generate in m
nD = 1000
Dvals = np.logspace(np.log10(x_min), np.log10(x_max), num=nD)
Nevaluations = 100
for name, ax in axes.items():
    ax.title.set_text(name)
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.set_ylabel("$\\mathregular{N_{>D}}$")
    ax.set_xlabel("Diameter (km)")
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)
    ax.yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
    ax.yaxis.set_minor_locator(
        ticker.LogLocator(base=10.0, subs=np.arange(2, 10), numticks=100)
    )
    ax.xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
    ax.xaxis.set_minor_locator(
        ticker.LogLocator(base=10.0, subs=np.arange(2, 10), numticks=100)
    )

    # Plot each sampled population Monte Carlo style
    for i in range(Nevaluations):
        Dsampled, _ = production[name].sample(
            age=age, diameter_range=diameter_range, area=area, return_age=False
        )
        Dsampled = np.sort(Dsampled)[::-1]
        Nsampled = range(1, len(Dsampled) + 1)
        ax.plot(
            Dsampled * 1e-3,
            Nsampled,
            "-",
            color="cornflowerblue",
            linewidth=2.0,
            zorder=50,
            alpha=0.2,
        )

    # Plot the production function (the expected values)
    Nvals = production[name].function(diameter=Dvals * 1e3, age=age)
    Nvals *= area  # convert from per unit area to total number
    ax.plot(Dvals, Nvals, "-", color="black", linewidth=1.0, zorder=50)

plt.tick_params(axis="y", which="minor")

plt.tight_layout()
plt.show()



import unittest

from numpy.random import default_rng

from cratermaker import Crater, Target


class TestCrater(unittest.TestCase):
    def test_crater_initialization_with_diameter(self):
        final_diameter = 1000
        crater = Crater.maker(final_diameter=final_diameter)
        self.assertEqual(crater.final_diameter, final_diameter)
        self.assertEqual(crater.final_radius, final_diameter / 2)
        self.assertIsNotNone(crater.transient_diameter)
        self.assertIsInstance(crater.transient_diameter, float)

    def test_crater_initialization_with_radius(self):
        final_radius = 500
        crater = Crater.maker(final_radius=final_radius)
        self.assertEqual(crater.final_radius, final_radius)
        self.assertEqual(crater.final_diameter, final_radius * 2)

    def test_crater_initialization_with_transient_diameter(self):
        transient_diameter = 800
        crater = Crater.maker(transient_diameter=transient_diameter)
        self.assertEqual(crater.transient_diameter, transient_diameter)
        self.assertGreater(crater.final_diameter, transient_diameter)

    def test_crater_initialization_with_transient_radius(self):
        transient_radius = 400
        crater = Crater.maker(transient_radius=transient_radius)
        self.assertEqual(crater.transient_radius, transient_radius)

    def test_invalid_negative_crater_values(self):
        with self.assertRaises(ValueError):
            Crater.maker(final_diameter=-1000)

    def test_invalid_crater_combinations(self):
        with self.assertRaises(ValueError):
            Crater.maker(final_diameter=1000, final_radius=500)

    def test_non_default_target_and_rng(self):
        rng = default_rng()
        target = Target("Mars")
        crater = Crater.maker(final_diameter=1000, target=target, rng=rng)
        self.assertEqual(crater.final_diameter, 1000)

    def test_invalid_target_or_rng_type(self):
        with self.assertRaises(ValueError):
            Crater.maker(final_diameter=1000, target="invalid_target")

    def test_projectile_initialization_with_diameter(self):
        diameter = 1000
        crater = Crater.maker(projectile_diameter=diameter, projectile_velocity=20e3)
        self.assertEqual(crater.projectile_diameter, diameter)
        self.assertEqual(crater.projectile_radius, diameter / 2)

    def test_projectile_initialization_with_radius(self):
        radius = 500
        crater = Crater.maker(projectile_radius=radius, projectile_velocity=20e3)
        self.assertEqual(crater.projectile_radius, radius)
        self.assertEqual(crater.projectile_diameter, radius * 2)

    def test_invalid_negative_projectile_values(self):
        with self.assertRaises(ValueError):
            Crater.maker(projectile_diameter=-1000, projectile_velocity=20e3)

    def test_invalid_projectile_combinations(self):
        with self.assertRaises(ValueError):
            Crater.maker(
                projectile_diameter=1000,
                projectile_radius=500,
                projectile_velocity=20e3,
            )
        with self.assertRaises(ValueError):
            Crater.maker(projectile_diameter=1000, final_diameter=10000)

    def test_non_default_target_and_rng_crater(self):
        rng = default_rng()
        crater = Crater.maker(
            projectile_diameter=1000, rng=rng, projectile_mean_velocity=10e3
        )
        self.assertEqual(crater.projectile_diameter, 1000)

    def test_invalid_target_or_rng_type_projectile(self):
        with self.assertRaises(ValueError):
            Crater.maker(
                projectile_diameter=1000,
                target="invalid_target",
                projectile_mean_velocity=10e3,
            )

    def test_argument_override(self):
        crater = Crater.maker(final_diameter=1000, location=(0.0, 0.0))
        self.assertEqual(crater.location, (0.0, 0.0))


if __name__ == "__main__":
    unittest.main()



"""
Create a shaded topographic representation of a crater
======================================================

.. rubric:: By David Minton

This example showcases how to create a crater and ejecta profile using the "simplemoon" morphology model from the Cratermaker and visual its topography. This will mimic how CTEM generates a test crater, though it is much simpler to run than that venerable old Fortran-based beast of a code!. The crater is created with a radius of 1 km. The hill shade uses the same settings that CTEM uses.

"""

import numpy as np

from cratermaker import Crater, Morphology

# Create morphology and crater objects
morphology = Morphology.maker("simplemoon")
crater = Crater.maker(final_radius=1.0e3)

# Generate 1000x1000 grid centered at (0, 0)
gridsize = 1000
extent = 5e3  # +/- extent in meters
pix = 2 * extent / gridsize  # pixel size in meters
x = np.linspace(-extent, extent, gridsize)
y = np.linspace(-extent, extent, gridsize)
xx, yy = np.meshgrid(x, y)

# Compute radial distance from center
r = np.sqrt(xx**2 + yy**2)

# Compute crater and ejecta profiles
crater_elevation = morphology.crater_profile(crater, r)
ejecta_elevation = morphology.ejecta_profile(crater, r)

# Combine into a total DEM
dem = crater_elevation + ejecta_elevation


def plot_hillshade(dem, pix):
    import matplotlib.pyplot as plt
    from matplotlib.colors import LightSource

    dpi = 300
    gridsize = dem.shape[0]
    # Create hillshade
    azimuth = 300.0  # user['azimuth']
    solar_angle = 20.0  # user['solar_angle']
    height = gridsize / dpi
    width = gridsize / dpi
    fig = plt.figure(figsize=(width, height), dpi=dpi)
    ax = plt.axes([0, 0, 1, 1])

    ls = LightSource(azdeg=azimuth, altdeg=solar_angle)
    hillshade = ls.hillshade(dem, dx=pix, dy=pix, fraction=1)

    # Plot the shaded relief
    ax.imshow(
        hillshade,
        interpolation="nearest",
        cmap="gray",
        vmin=0.0,
        vmax=1.0,
        extent=(-extent, extent, -extent, extent),
    )
    plt.axis("off")
    plt.show()


plot_hillshade(dem, pix)



from __future__ import annotations

import hashlib
import math
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

from numpy.random import Generator

from ..utils.general_utils import format_large_units
from .base import CratermakerBase

if TYPE_CHECKING:
    from ..components.projectile import Projectile
    from ..components.scaling import Scaling
    from ..components.target import Target


@dataclass(frozen=True, slots=True)
class Crater:
    final_diameter: float | None = None
    transient_diameter: float | None = None
    projectile_diameter: float | None = None
    projectile_velocity: float | None = None
    projectile_angle: float | None = None
    projectile_direction: float | None = None
    projectile_density: float | None = None
    location: tuple[float, float] | None = None
    morphology_type: str | None = None
    age: float | None = None

    def __str__(self):
        if self.age is None:
            agetext = "Not set"
        else:
            agetext = f"{format_large_units(self.age, quantity='time')}"
        return (
            f"final_diameter: {format_large_units(self.final_diameter, quantity='length')}\n"
            f"transient_diameter: {format_large_units(self.transient_diameter, quantity='length')}\n"
            f"projectile_diameter: {format_large_units(self.projectile_diameter, quantity='length')}\n"
            f"projectile_mass: {self.projectile_mass:.4e} kg\n"
            f"projectile_density: {self.projectile_density:.0f} kg/m^3\n"
            f"projectile_velocity: {format_large_units(self.projectile_velocity, quantity='velocity')}\n"
            f"projectile_angle: {self.projectile_angle:.1f}°\n"
            f"projectile_direction: {self.projectile_direction:.1f}°\n"
            f"location (lon,lat): ({self.location[0]:.4f}°, {self.location[1]:.4f}°)\n"
            f"morphology_type: {self.morphology_type}\n"
            f"age: {agetext}"
        )

    @property
    def _id(self):
        """
        The hash id of the crater. This is used as a unique identifier for the crater.
        """
        # Ensure deterministic field order by sorting keys (optional but safer)
        data = asdict(self)
        combined = ":".join(str(data[k]) for k in sorted(data))
        return hashlib.sha256(combined.encode()).hexdigest()

    @property
    def final_radius(self) -> float | None:
        """Final radius of the crater in meters."""
        return self.final_diameter / 2.0 if self.final_diameter is not None else None

    @property
    def transient_radius(self) -> float | None:
        """Transient radius of the crater in meters."""
        return (
            self.transient_diameter / 2.0
            if self.transient_diameter is not None
            else None
        )

    @property
    def projectile_radius(self) -> float | None:
        """Projectile radius in meters."""
        return (
            self.projectile_diameter / 2.0
            if self.projectile_diameter is not None
            else None
        )

    @property
    def projectile_mass(self) -> float | None:
        """Projectile mass in kilograms."""
        if self.projectile_density is not None and self.projectile_radius is not None:
            return (
                (4.0 / 3.0)
                * math.pi
                * self.projectile_radius**3
                * self.projectile_density
            )
        return None

    @property
    def projectile_vertical_velocity(self) -> float | None:
        """Projectile vertical velocity in m/s."""
        if self.projectile_velocity is not None and self.projectile_angle is not None:
            return self.projectile_velocity * math.sin(
                math.radians(self.projectile_angle)
            )
        return None

    @classmethod
    def maker(
        cls: type[Crater],
        crater: Crater | None = None,
        scaling: str | Scaling | None = None,
        target: str | Target | None = None,
        projectile: str | Projectile | None = None,
        final_diameter: float | None = None,
        final_radius: float | None = None,
        transient_diameter: float | None = None,
        transient_radius: float | None = None,
        projectile_diameter: float | None = None,
        projectile_radius: float | None = None,
        projectile_mass: float | None = None,
        projectile_density=None,
        projectile_velocity: float | None = None,
        projectile_mean_velocity: float | None = None,
        projectile_vertical_velocity=None,
        projectile_angle: float | None = None,
        projectile_direction: float | None = None,
        projectile_location: tuple[float, float] | None = None,
        location: tuple[float, float] | None = None,
        age: float | None = None,
        simdir: str | Path | None = None,
        rng: Generator = None,
        rng_seed: str | int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        """
        Create a Crater object with the given parameters.

        Parameters
        ----------
        crater : Crater, optional
            A Crater object to copy parameters from.
        scaling : str or Scaling, optional
            A string key or instance of a scaling model. If none provided, a default will be used
        target : str or Target, optional
            A string key or instance of a target model. If none provided, a default will be used.
        projectile : str or Projectile, optional
            A string key or instance of a projectile model. If none provided, a default will be used.
        final_diameter : float, optional
            The final diameter of the crater in meters.
        final_radius : float, optional
            The final radius of the crater in meters.
        transient_diameter : float, optional
            The transient diameter of the crater in meters.
        transient_radius : float, optional
            The transient radius of the crater in meters.
        projectile_diameter : float, optional
            The diameter of the projectile in meters.
        projectile_radius : float, optional
            The radius of the projectile in meters.
        projectile_mass : float, optional
            The mass of the projectile in kilograms.
        projectile_density : float, optional
            The density of the projectile in kg/m^3. If not provided, the target's density is used.
        projectile_velocity : float, optional
            The total impact velocity of the projectile in m/s.
        projectile_mean_velocity : float, optional
            The mean velocity from which to sample a projectile velocity.
        projectile_vertical_velocity : float, optional
            The vertical component of the velocity in m/s.
        projectile_angle : float, optional
            The impact angle in degrees (0-90).
        projectile_direction : float, optional
            The direction of the impact in degrees (0-360).
        projectile_location : tuple of float, optional
            The (longitude, latitude) location of the projectile impact. This is equivalent to `location`, which takes precedence
        location : tuple of float, optional
            The (longitude, latitude) location of the crater.
        age : float, optional
            The age of the crater in Myr.
        simdir : str | Path
            The main project simulation directory. Defaults to the current working directory if None.
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments for subclasses.

        Returns
        -------
        Crater
            A frozen Crater dataclass with derived attributes.

        Notes
        -----
        - Exactly one of the following must be provided: `final_diameter`, `final_radius`, `transient_diameter`, `transient_radius`, `projectile_diameter`, `projectile_radius`, or `projectile_mass`.
        - Velocity may be specified in one of these ways:
        - `projectile_mean_velocity` alone (samples a velocity)
        - Any two of (`projectile_velocity`, `projectile_vertical_velocity`, `projectile_angle`). the third is inferred.
        - `projectile` is mutually exclusive with velocity-related inputs; if provided, it overrides velocity, angle, direction, and density unless explicitly set.
        - The `scaling`, and `rng` models are required for scaling and density inference, but are not stored in the returned Crater object.
        """
        from ..components.projectile import Projectile
        from ..components.scaling import Scaling
        from ..components.target import Target

        # Validate that mutually exclusive arguments hve not been passed
        size_inputs = {
            "final_diameter": final_diameter,
            "final_radius": final_radius,
            "transient_diameter": transient_diameter,
            "transient_radius": transient_radius,
            "projectile_diameter": projectile_diameter,
            "projectile_radius": projectile_radius,
            "projectile_mass": projectile_mass,
        }

        velocity_inputs = {
            "projectile_velocity": projectile_velocity,
            "projectile_vertical_velocity": projectile_vertical_velocity,
            "projectile_angle": projectile_angle,
        }

        if location is None:
            location = projectile_location

        args = {
            "final_diameter": final_diameter,
            "transient_diameter": transient_diameter,
            "projectile_diameter": projectile_diameter,
            "projectile_density": projectile_density,
            "projectile_velocity": projectile_velocity,
            "projectile_angle": projectile_angle,
            "projectile_direction": projectile_direction,
            "morphology_type": "Not Set",
            "location": location,
            "age": age,
        }

        n_velocity_inputs = sum(x is not None for x in velocity_inputs.values())
        if n_velocity_inputs > 2:
            raise ValueError(
                f"Only two of {', '.join(k for k, v in velocity_inputs.items() if v is not None)} may be set."
            )

        if projectile_mean_velocity is not None:
            if (
                projectile_velocity is not None
                or projectile_vertical_velocity is not None
            ):
                raise ValueError(
                    "projectile_mean_velocity cannot be used with projectile_velocity or projectile_vertical_velocity"
                )

        n_size_inputs = sum(v is not None for v in size_inputs.values())

        # Process input crater object if provided
        if crater is not None:
            if not isinstance(crater, Crater):
                raise TypeError("crater must be a Crater object.")
            old_parameters = {}
            for field in cls.__dataclass_fields__:
                if field in locals() and locals()[field] is None:
                    old_parameters[field] = getattr(crater, field)
            if (
                n_size_inputs == 0
            ):  # The user has not passed any size parameters, so we will use the final diameter from the crater object
                for field in size_inputs:
                    if field != "final_diameter":
                        old_parameters.pop(field, None)
                n_size_inputs = 1  # Make sure we don't trigger the error below
            else:  # The user is passing a size parameter, so we cannot use any of the values from the crater object
                for field in size_inputs:
                    old_parameters.pop(field, None)
            if projectile_mean_velocity is None:
                # Be sure to keep only two velocitity components, with a preference for angle over vertical velocity
                if n_velocity_inputs == 0:
                    old_parameters.pop("projectile_vertical_velocity", None)
                elif n_velocity_inputs == 1:
                    for k, v in velocity_inputs.items():
                        if v is not None:
                            old_parameters.pop(k, None)
                            if k == "projectile_velocity" or k == "projectile_angle":
                                old_parameters.pop("projectile_vertical_velocity", None)
                            elif k == "projectile_vertical_velocity":
                                old_parameters.pop("projectile_velocity", None)
                elif n_velocity_inputs == 2:
                    for field in velocity_inputs:
                        old_parameters.pop(field, None)
                # Make sure we don't override the projectile_velocity and projectile_angle
                n_velocity_inputs = 2

            # Now set the local arguments to be what's left from the old_parameters
            for field in old_parameters:
                if field in args and args[field] is None:
                    args[field] = old_parameters[field]

        if n_size_inputs != 1:
            raise ValueError(
                f"Exactly one of {', '.join(k for k, v in size_inputs.items() if v is not None)} must be set."
            )

        # --- Normalize RNG, rng_seed, simdir using CratermakerBase ---
        argproc = CratermakerBase(
            simdir=simdir, rng=rng, rng_seed=rng_seed, rng_state=rng_state
        )
        if scaling is not None and isinstance(scaling, Scaling):
            if projectile is None:
                projectile = scaling.projectile
            if target is None:
                target = scaling.target

        target = Target.maker(target, **vars(argproc.common_args), **kwargs)

        projectile_args = {
            "velocity": args["projectile_velocity"],
            "density": args["projectile_density"],
            "angle": args["projectile_angle"],
            "direction": args["projectile_direction"],
            "location": args["location"],
        }
        projectile = Projectile.maker(
            projectile,
            target=target,
            **vars(argproc.common_args),
            **kwargs,
        ).new_projectile(
            **projectile_args,
        )

        scaling = Scaling.maker(
            scaling,
            target=target,
            projectile=projectile,
            **vars(argproc.common_args),
            **kwargs,
        )
        projectile = scaling.projectile
        scaling.recompute()
        target = scaling.target
        pv = projectile.velocity
        pvv = projectile.vertical_velocity
        pang = projectile.angle
        pdir = projectile.direction
        prho = projectile.density
        location = projectile.location

        # --- Ensure velocity/angle are all set ---
        n_set = sum(x is not None for x in [pv, pvv, pang])
        if n_set != 3:
            raise ValueError("Not enough information to infer a projectile velocity.")

        # --- Compute derived quantities ---
        fd = args["final_diameter"]
        fr = final_radius
        td = args["transient_diameter"]
        tr = transient_radius
        pd = args["projectile_diameter"]
        pr = projectile_radius
        pm = projectile_mass
        mt = None

        if pr is not None:
            pd = 2 * pr
        if fr is not None:
            fd = 2 * fr
        if tr is not None:
            td = 2 * tr

        if fd is not None:
            td, mt = scaling.final_to_transient(fd)
            pd = scaling.transient_to_projectile(td)
        elif td is not None:
            fd, mt = scaling.transient_to_final(td)
            pd = scaling.transient_to_projectile(td)
        elif pd is not None:
            td = scaling.projectile_to_transient(pd)
            fd, mt = scaling.transient_to_final(td)
        elif pm is not None:
            pr = ((3.0 * pm) / (4.0 * math.pi * prho)) ** (1.0 / 3.0)
            pd = 2.0 * pr
            td = scaling.projectile_to_transient(pd)
            fd, mt = scaling.transient_to_final(td)

        pr = pd / 2
        tr = td / 2
        fr = fd / 2
        pm = (4.0 / 3.0) * math.pi * pr**3 * prho

        # Assemble final arguments
        args = {
            "final_diameter": float(fd) if fd is not None else None,
            "transient_diameter": float(td) if td is not None else None,
            "projectile_diameter": float(pd) if pd is not None else None,
            "projectile_density": float(prho) if prho is not None else None,
            "projectile_velocity": float(pv) if pv is not None else None,
            "projectile_angle": float(pang) if pang is not None else None,
            "projectile_direction": float(pdir) if pdir is not None else None,
            "location": (float(location[0]), float(location[1])),
            "morphology_type": str(mt) if mt is not None else None,
            "age": float(age) if age is not None else None,
        }
        return cls(**args)



from pathlib import Path
from typing import Any, Callable
from warnings import warn

import numpy as np
import yaml
from numpy.typing import ArrayLike

from cratermaker.constants import FloatLike


class Parameter(property):
    """
    A property descriptor that tracks user-defined properties.  This class is a subclass of the built-in property class and is used
    to create properties in a class that can be set and retrieved. It also tracks whether the property has been set by the user,
    allowing for parameters to be exported to a YAML configuration file.
    """

    def __init__(self, fget, fset=None, fdel=None, doc=None):
        super().__init__(fget, fset, fdel, doc)
        self.name = fget.__name__

    def setter(self, fset):
        def wrapped(instance, value):
            if not hasattr(instance, "_user_defined"):
                instance._user_defined = set()
            instance._user_defined.add(self.name)
            fset(instance, value)

        return Parameter(self.fget, wrapped, self.fdel, self.__doc__)


def parameter(fget=None):
    """
    A decorator to mark a property as a user-settable parameter.
    Can be used with or without parentheses.
    """
    if fget is None:

        def decorator(fget):
            return Parameter(fget)

        return decorator
    else:
        return Parameter(fget)


def _set_properties(
    obj,
    catalogue: dict | None = None,
    key: str | None = None,
    config_file: str | Path | None = None,
    **kwargs: Any,
):
    """
    Set properties of a simulation object from various sources.

    This function sets the properties of a simulation object based on the provided arguments.
    Properties can be read from a YAML file, a pre-defined catalogue, or directly passed as keyword arguments.

    Parameter
    ----------
    obj : object
        The simulation object whose properties are to be set.
    catalogue : dict, optional
        A dictionary representing a catalogue of properties. It must be in the form of a nested dict. If provided, it will be used to set properties.
    key : str, optional
        The key to look up in the catalogue. It must be provided if the catalogue is provided.
    config_file : str or Path, optional
        The path to a YAML file containing properties. If provided, it will be used to set properties.
    **kwargs : dict
        Keyword arguments that can include 'config_file', 'catalogue', and other direct property settings.

    Returns
    -------
    matched : dict
        A dictionary of properties that were successfully set on the object.
    unmatched : dict
        A dictionary of properties that were not set, either due to being None or not matching any known properties.

    Notes
    -----
    The order of property precedence is:
    1. Direct keyword arguments (kwargs).
    2. Pre-defined catalogue (specified by 'catalogue' key in kwargs).
    3. YAML file (specified by 'config_file' key in kwargs).
    Properties set by kwargs override those set by 'catalogue' or 'config_file'.
    """

    def _set_properties_from_arguments(obj, **kwargs):
        matched = {}
        unmatched = {}
        cls = type(obj)
        for key, value in kwargs.items():
            if value is None:
                continue
            param = getattr(cls, key, None)
            if (
                isinstance(param, (property, Parameter))
                and getattr(param, "fset", None) is not None
            ):
                setattr(obj, key, value)
                matched[key] = value
            else:
                unmatched[key] = value
        return matched, unmatched

    def _set_properties_from_catalogue(obj, catalogue, key, **kwargs):
        if "catalogue_key" in dir(obj):
            catalogue_key = getattr(obj, "catalogue_key")
        else:
            raise ValueError(
                "The object does not have a catalogue_key property, and therefore is not set up to receive catalogue entries."
            )
        if catalogue_key in kwargs:
            key = kwargs.pop(catalogue_key)

        if not isinstance(catalogue, dict):
            raise ValueError("Catalogue must be a dictionary")

        for k, v in catalogue.items():
            if not isinstance(v, dict):
                raise ValueError(
                    f"Value for key '{k}' in catalogue must be a dictionary"
                )

        if key not in catalogue:
            return {}, {}

        properties = catalogue.get(key)
        properties.update({catalogue_key: key})
        # Remove any items in kwargs that are already in properties
        for k in properties.keys():
            if k in kwargs:
                del kwargs[k]
        if properties:  # A match was found to the catalogue
            matched, unmatched = _set_properties_from_arguments(
                obj, **properties, **kwargs
            )
            properties.pop(
                catalogue_key
            )  # Make sure that the catlogue key doesn't stay in the properties
        return matched, unmatched

    def _set_properties_from_file(obj, config_file, key=None, **kwargs):
        try:
            with open(config_file, "r") as f:
                properties = yaml.safe_load(f)
        except Exception as e:
            warn(f"Could not read the file {config_file}.\n{e}", RuntimeWarning)
            return {}, {}
        merged = {**properties, **{k: v for k, v in kwargs.items() if v is not None}}
        if key is None:
            matched, unmatched = _set_properties_from_arguments(obj, **merged)
        else:
            if key not in properties:
                raise ValueError(f"Key '{key}' not found in the file '{config_file}'.")
            matched, unmatched = _set_properties_from_catalogue(
                obj, key=key, catalogue=properties, **kwargs
            )
        return matched, unmatched

    matched = {}
    unmatched = {}
    if config_file:
        m, u = _set_properties_from_file(
            obj, config_file=config_file, key=key, **kwargs
        )
        matched.update(m)
        unmatched.update(u)

    if catalogue:
        m, u = _set_properties_from_catalogue(
            obj, catalogue=catalogue, key=key, **kwargs
        )
        matched.update(m)
        unmatched.update(u)

    m, u = _set_properties_from_arguments(obj, **kwargs)
    matched.update(m)
    unmatched.update(u)

    # if there are any keys in unmatched that are also present in matched, remove them from unmatched
    for key in matched.keys():
        if key in unmatched:
            del unmatched[key]

    return matched, unmatched


def _create_catalogue(header, values):
    """
    Create and return a catalogue of properties or items based on the given inputs.

    This function generates a catalogue, which could be a collection of properties, configurations,
    or any other set of items, based on the provided arguments.

    Parameter
    ----------
    args : various
        The arguments that determine the contents of the catalogue. The type and number of arguments
        can vary based on the intended use of the catalogue.

    Returns
    -------
    catalogue_type
        A catalogue of items or properties. The exact type of this catalogue (e.g., dict, list, custom object)
        depends on the implementation.

    Notes
    -----
    The catalogues built by this function are the built-in catalogues for material properties and target bodie
    """
    # Create the catalogue dictionary using the class variables
    catalogue = {tab[0]: dict(zip(header, tab)) for tab in values}

    # Remove the first key from each dictionary in the catalogue
    for k in list(catalogue):
        del catalogue[k][header[0]]

    return catalogue


def normalize_coords(location: tuple[FloatLike, FloatLike]) -> tuple[float, float]:
    """
    Normalize geographic coordinates to ensure longitude is within [-180, 180) degrees
    and latitude within [-90, 90] degrees.

    This function takes a tuple of longitude and latitude values in degrees, normalizes
    them to the specified ranges, and handles cases where latitude values exceed the
    polar extremes, adjusting both latitude and longitude accordingly.

    Parameter
    ----------
    location : tuple
        A tuple containing two elements: (longitude, latitude) in degrees.
        Longitude and latitude can be any float values.

    Returns
    -------
    tuple
        A tuple of two elements: (normalized_longitude, normalized_latitude).
        The normalized longitude is in the range [-180, 180) degrees, and the
        normalized latitude is in the range [-90, 90] degrees.

    Notes
    -----
    - The longitude is normalized using a modulo operation with 360 degrees and then adjusted to the range [-180, 180).
    - Latitude values beyond 90 or below -90 degrees are adjusted by reflecting them within the range and flipping the longitude by 180 degrees, then re-normalizing it to the [-180, 180) range.

    Examples
    --------
    >>> normalize_coords((370, 95))
    (10.0, 85.0)

    >>> normalize_coords((-185, -100))
    (-5.0, 80.0)
    """
    lon, lat = location

    # Normalize longitude to be within [-180, 180)
    normalized_lon = ((lon + 180) % 360) - 180

    # Normalize latitude
    if lat > 90:
        normalized_lat = 180 - lat
        normalized_lon = lon - 180  # Flip the longitude
    elif lat < -90:
        normalized_lat = -180 - lat
        normalized_lon = lon - 180  # Flip the longitude
    else:
        normalized_lat = lat

    # Ensure latitude is within the range [-90, 90] after adjustments
    normalized_lat = np.clip(normalized_lat, -90, 90)

    return float(normalized_lon), float(normalized_lat)


def validate_and_normalize_location(location):
    """
    Validate and normalize a given location into a standard structured format.

    This function checks the input location data and converts it into a
    consistent structured array format if it is a valid location representation.
    Valid formats for location include a tuple, a dictionary, or a structured
    array with longitude ('lon') and latitude ('lat').

    Parameters
    ----------
    location : tuple, dict, ArrayLike
        The input location data. It can be:
        - A tuple, list, or array with two elements (longitude, latitude).
        - A dictionary with keys 'lon' and 'lat'.
        - A structured numpy array with 'lon' and 'lat' fields.

    Returns
    -------
    tuple
        longitude and latitude as a tuple of floats.

    Raises
    ------
    ValueError
        If the input does not conform to one of the expected formats for location data.

    Examples
    --------
    >>> validate_and_normalize_location((370, 95))
    (10.0, 85.0))

    >>> validate_and_normalize_location({'lat': 45.0, 'lon': 120.0})
    (-120., 45.)

    >>> validate_and_normalize_location(np.array([(-120.0, 45.0)], dtype=[('lon', 'f8'), ('lat', 'f8')]))
    (-120., 45.)

    """
    # Check if it's already a tuple

    if isinstance(location, np.ndarray) and location.dtype.names == ("lon", "lat"):
        return normalize_coords((location[0], location[1]))

    if isinstance(location, np.ndarray) and location.dtype.names == ("lat", "lon"):
        return normalize_coords((location[1], location[0]))

    if isinstance(location, (tuple, list, np.ndarray)) and len(location) == 2:
        return normalize_coords(location)

    # Check if it's a dictionary with 'lon' and 'lat' keys
    if isinstance(location, dict):
        if "lon" in location and "lat" in location:
            return normalize_coords((location["lon"], location["lat"]))

    if len(location) == 2:
        return normalize_coords((location[0], location[1]))

    raise ValueError(
        "location must a tuple, list, or ArrayLike of len==2, a dict with 'lon' and 'lat', or a structured array with 'lon' and 'lat' names"
    )


def R_to_CSFD(
    R: Callable[[FloatLike | ArrayLike], FloatLike | ArrayLike],
    D: FloatLike | ArrayLike,
    Dlim: FloatLike = 1e6,
    *args: Any,
) -> FloatLike | ArrayLike:
    """
    Convert R values to cumulative N values for a given D using the R-plot function.

    Parameter
    ----------
    R : R = f(D)
        A function that computes R given D.
    D : FloatLike or ArrayLike
        diameter in units of km.
    Dlim : FloatLike
        Upper limit on the diameter over which to evaluate the integral
    args : Any
        Additional arguments to pass to the R function

    Returns
    -------
    float or ArrayLike
        The cumulative number of craters greater than D in diameter.
    """

    def _R_to_CSFD_scalar(R, D, Dlim, *args):
        # Helper function to integrate the R function
        def integrand(D):
            return R(D, *args) / D**3  # This is dN/dD

        N = 0.0
        D_i = D
        while D_i < Dlim:
            D_next = D_i * np.sqrt(2.0)
            D_mid = (D_i + D_next) / 2  # Mid-point of the bin
            bin_width = D_next - D_i
            R_value = integrand(D_mid)
            N += R_value * bin_width
            D_i = D_next  # Move to the next bin

        return N

    return (
        _R_to_CSFD_scalar(R, D, Dlim, *args)
        if np.isscalar(D)
        else np.vectorize(_R_to_CSFD_scalar)(R, D, Dlim, *args)
    )


def format_large_units(
    value: float, threshold: float = 1000.0, quantity: str = "length"
) -> str:
    """
    Format a value and automatically shift units based on threshold.
    """
    if quantity == "length":
        units = ["m", "km"]
    elif quantity == "velocity":
        units = ["m/s", "km/s"]
    elif quantity == "time":
        units = ["My", "Gy"]
    elif quantity == "pressure":
        units = ["Pa", "kPa", "MPa", "GPa"]

    if value is None:
        return "N/A"

    unit_index = 0
    while unit_index + 1 < len(units) and value >= threshold:
        value /= threshold
        unit_index += 1

    if value >= 100:
        fmt = "{:.0f} {}"
    elif value >= 10:
        fmt = "{:.1f} {}"
    elif value >= 1:
        fmt = "{:.2f} {}"
    else:
        fmt = "{:.3g} {}"
    return fmt.format(value, units[unit_index])



"""
Topographic Diffusion
=====================

.. rubric:: By Dennise Valadez and David Minton

This example demonstrates how to use the :meth:`apply_diffusion` method in the :ref:`Surface <ug-surface>` class to model topographic diffusion. In this example, we will simulate the change in elevation over time of a hill with a Gaussian profile. This example has an analytical solution, which we will compare against the numerical solution provided by the diffusion method.

For this example we will use the "hireslocal" surface type, which will give us a high-resolution local region of the surface that will be approximately a flat surface.

"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import griddata

from cratermaker import Surface


def analytical_elevation(r, h0, sigma, kappa, t):
    """Analytical solution for elevation profile after diffusion."""
    variance_t = sigma**2 + 2 * kappa * t
    return h0 * (sigma**2 / variance_t) * np.exp(-(r**2) / (2 * variance_t))


# We will set up a surface with a 2 km radius high resolution region at 10 m/pix. The location is arbitrary, but we will set it to (0, 0) for simplicity (the equator/prime meridian).
local_radius = 2000.0
pix = 10.0
local_location = (0, 0)
superdomain_scale_factor = 10000

# The initial elevation will be a Gaussian hill with a height of 100 m and a standard deviation of 200 m.
h0 = 100.0
sigma = 200.0

# The diffusivity κ will be set to 5000 m² per time step (in arbitrary time units), and we will run the simulation for a total time of 10 time steps.
kappa = 5000.0
nsteps = 10
total_time = 10.0  # total simulated time in arbitrary units
dt = total_time / nsteps

surface = Surface.maker(
    "hireslocal",
    local_location=local_location,
    pix=pix,
    local_radius=local_radius,
    superdomain_scale_factor=superdomain_scale_factor,
)

# Generate the initial elevation profile and apply it to the surface
r = surface.local.face_distance
h = analytical_elevation(r, h0, sigma, kappa, 0.0)
isort = np.argsort(r)
surface.local.update_elevation(h)

# We will loop over the number of steps, applying diffusion at each step and plotting the results, starting from the initial condition
for i in range(nsteps + 1):
    t = i * dt
    label_num = "Initial" if i == 0 else f"Step {i}"
    label_ana = "Analytical" if i == nsteps else None
    h_numerical = surface.local.face_elevation
    h_analytical = analytical_elevation(r, h0, sigma, kappa, t)
    plt.plot(r[isort], h_numerical[isort], label=label_num)
    plt.plot(r[isort], h_analytical[isort], linestyle="--", color="black", alpha=0.5, label=label_ana)
    # Compare difference between analytical and numerical statistically
    rmse = np.sqrt(np.mean((h_numerical - h_analytical) ** 2))
    nrmse = rmse / np.ptp(h_analytical)
    print(f"{label_num:<7}: Normalized root mean square error = {nrmse:.4%}")
    surface.local.apply_diffusion(kappa)

plt.xlabel("Distance (m)")
plt.ylabel("Elevation (m)")
plt.xlim(0, local_radius)

plt.title("Elevation Profile with Topographic Diffusion")
plt.legend()
plt.tight_layout()
plt.show()



"""
Plot a distribution of random locations
=======================================

.. rubric:: By David Minton and Dennise Valadez

This example demonstrates the generation of random impact locations.

"""

import matplotlib.pyplot as plt
import numpy as np

from cratermaker import Projectile

# Create a projectile
proj = Projectile(mean_velocity=5000, density=3000)

# Using projectile for sampling data
size = 10000
locations = [proj.new_projectile().location for _ in range(size)]

# Obtain longitudes and latitudes from projectile - but this is not accurate and doesnt work
lons = np.array([loc[0] for loc in locations])
lats = np.array([loc[1] for loc in locations])

# Bins
bins = 50

# Longitude histogram
observed_counts_lon, bins_lon_deg = np.histogram(lons, bins=bins, range=(-180, 180.0))
expected_count_lon = size // bins

# Latitude histogram
observed_counts_lat, bins_lat_deg = np.histogram(lats, bins=bins, range=(-90, 90))
bins_lat = np.deg2rad(bins_lat_deg)
area_ratio = np.sin(bins_lat[1:]) - np.sin(bins_lat[:-1])
total_area = np.sin(np.pi / 2) - np.sin(-np.pi / 2)
expected_count_lat = size * area_ratio / total_area

# Bar widths
bar_width_lon = np.diff(bins_lon_deg)
bar_width_lat = np.diff(bins_lat_deg)

# Plotting
fig, axs = plt.subplots(2, 1, figsize=(8, 6))

# Longitude plot
axs[0].bar(
    bins_lon_deg[:-1],
    observed_counts_lon,
    width=bar_width_lon,
    align="edge",
    label="Observed",
)
axs[0].plot(
    bins_lon_deg[:-1],
    [expected_count_lon] * len(bins_lon_deg[:-1]),
    color="red",
    label="Expected",
)
axs[0].set_xlabel("Longitude (deg)")
axs[0].set_ylabel("Number")
axs[0].legend()
axs[0].set_title("Number vs Longitude")

# Latitude plot
axs[1].bar(
    bins_lat_deg[:-1],
    observed_counts_lat,
    width=bar_width_lat,
    align="edge",
    alpha=0.5,
    label="Observed",
)
axs[1].plot(bins_lat_deg[:-1], expected_count_lat, label="Expected", color="red")
axs[1].set_xlabel("Latitude (deg)")
axs[1].set_ylabel("Number")
axs[1].legend()
axs[1].set_title("Number vs Latitude")

plt.tight_layout()
plt.show()



from typing import Any

from numpy.random import Generator

from cratermaker.components.projectile import Projectile
from cratermaker.components.target import Target
from cratermaker.constants import FloatLike


@Projectile.register("generic")
class GenericProjectiles(Projectile):
    def __init__(
        self,
        sample: bool = True,
        mean_velocity: FloatLike | None = None,
        velocity: FloatLike | None = None,
        density: FloatLike | None = None,
        angle: FloatLike | None = None,
        direction: FloatLike | None = None,
        location: tuple[FloatLike, FloatLike] | None = None,
        target: Target | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        """
        An operations class for computing the projectile properties of a generic source population. The user is required to either set `mean_velocity` (if `sample==True`) or `velocity` (if `sample==False`).
        If `sample==True`, the impact velocities, angles, and directions are sampled from distributions. If `sample==False`, the impact velocities are set to the mean velocity, and `angle` and `direction` can also be set, but will default to set 90 degrees (vertical impact) and 0 degrees (due North`).

        Parameters
        ----------
        sample : bool
            Flag that determines whether to sample impact velocities, angles, and directions from distributions. If set to True, the `mean_velocity` argument is required. If set to False, the `velocity` argument is required.
        mean_velocity : float, optional
            The mean velocity of the projectile in m/s. Required if `sample` is True, ignored if `sample` is False.
        velocity : float | None
            The impact velocity in m/s. If `sample` is True, this value is ignored. If `sample` is False, this value is required.
        density : float, optional
            The density of the projectile in kg/m^3. Default is 1000 kg/m^3.
        angle : float, optional
            The impact angle in degrees. Default is 90.0 degrees (vertical impact) if `sample` is False. If `sample` is True, this value is ignored.
        direction : float | None
            The impact direction in degrees. Default is 0.0 degrees (due North) if `sample` is False. If `sample` is True, this value is ignored.`
        location : tuple[float, float] | None
            The impact location as a tuple of (longitude, latitude) in degrees. Default is (0.0, 0.0) (the equator and prime meridian) if `sample` is False. If `sample` is True, this value is ignored.
        target : Target or str.
            The name of the target body for the impact. Default is "Moon"
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.
        """

        # set the defaults
        density = 1000.0 if density is None else density
        angle = 90.0 if angle is None else angle
        direction = 0.0 if direction is None else direction
        location = (0.0, 0.0) if location is None else location
        super().__init__(
            sample=sample,
            mean_velocity=mean_velocity,
            velocity=velocity,
            density=density,
            angle=angle,
            direction=direction,
            location=location,
            target=target,
            rng=rng,
            rng_seed=rng_seed,
            rng_state=rng_state,
            **kwargs,
        )
        return



"""
Plot a distribution of random velocities given a mean
=====================================================

.. rubric:: By David Minton and Dennise Valadez

This example demonstrates the generation of random velocities using the :func:`get_random_velocity` function from the :ref:`montecarlo_utils <api-utils-montecarlo>` module. The expected distribution of imapct velocities is a modified Maxwell-Boltzmann distribution. The modification is related to the relationship between the mean velocity and the escape velocity. If vmean > vescape, then the distribution will adjusted so that the escape velocity is the minimum velocity by computing an encounter velocity then summing the encounter and escape velocities in quadrature. If vmean < vescape, then the distirbution will be a truncated maxwellian.

Here we sample 10000 velocities with a mean of 0.5 and 2x the escape velocity of the Moon, both with and without the escape velocity argument. Then we will compare a histogram of generated velocities with .

"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import maxwell

from cratermaker import Projectile, Target

# Use Moon as target
target = Target.maker("Moon")
vescape = target.escape_velocity
size = 10000

factors = np.array([0.75, 2.0])
vmean_values = factors * vescape
titles = [f"vmean = {float(factor):0.2f} $\\times$ vescape" for factor in factors]

fig, axs = plt.subplots(2, 1, figsize=(12, 6), sharex="col", sharey="row")

for col, vmean in enumerate(vmean_values):
    # Initialize projectiles with and without escape velocity
    proj = Projectile(mean_velocity=vmean, density=3000)

    velocities = np.array([proj.new_projectile().velocity for _ in range(size)])

    bins = np.linspace(0, vmean * 3 / vescape, 50)
    histogram, bins = np.histogram(velocities / vescape, bins=bins, density=True)
    bin_centers = 0.5 * (bins[1:] + bins[:-1])
    theoretical_dist = (
        maxwell.pdf(bin_centers * vescape, scale=vmean / np.sqrt(8 / np.pi)) * vescape
    )

    ax = axs[col]
    ax.bar(
        bin_centers,
        histogram,
        width=bins[1] - bins[0],
        alpha=0.5,
        label="Observed",
    )
    ax.plot(
        bin_centers,
        theoretical_dist,
        label="Maxwell-Boltzmann",
        color="red",
    )
    rms = np.sqrt(np.mean(velocities**2)) / vescape
    ax.axvline(rms, color="orange", linestyle="--", label="RMS Velocity")
    ax.annotate(
        f"RMS = {rms:.2f}",
        xy=(rms, ax.get_ylim()[1] * 0.65),
        xytext=(5, 0),
        textcoords="offset points",
        rotation=90,
        va="top",
        ha="left",
        fontsize=12,
    )
    if col == 0:
        ax.set_ylabel("Probability Density")
        ax.set_xlabel("$V / V_{{escape}}$")
        ax.legend()
        ax.set_title(titles[col])

plt.tight_layout()
plt.show()



"""
Crater scaling for various planetary surfaces
==============================================

.. rubric:: By Dennise Valadez and David Minton

This example demonstrates how to use the :ref:`Scaling <ug-scaling>` class to calculate the final crater diameter for different planetary bodies.  Normally, the scaling component would not be invoked manually like this. Instead, the scaling model is used internally by the :ref:`Crater <ug-crater>` component. This example demonstrates how the scaling model can be used on its own.

This example uses the default "montecarlo" scaling model, in which some of the scaling relationships are randomly varied for each computation. It also uses the default :ref:`Projectile <ug-projectile>` model, which draws projectile velocities and angles at random from distributions. Therefore, the results will be slightly different each time it is run.
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from cratermaker import Scaling

 
bodies = ["Mercury", "Venus", "Earth", "Moon", "Mars", "Ceres", "Vesta"]
proj_diameters = np.logspace(1, 4, 100)
markersize = 6
# Define different marker styles for each body
markerstyles = ['o', 'X', 'D']
morphology_types = ['simple', 'transitional', 'complex']
markerstyle_map = dict(zip(morphology_types, markerstyles))
colors = ["gray", "gold", "dodgerblue", "saddlebrown", "salmon", "powderblue", "mediumorchid" ]
color_map = dict(zip(bodies, colors))

fig, ax = plt.subplots(figsize=(9, 7))
ax.set_xscale("log")
ax.set_yscale("log")
ax.set_xlabel("Projectile Diameter [km]")
ax.set_ylabel("Final Crater Diameter [km]")
ax.set_title("Projectile to Crater Scaling for Mars, Venus, Mercury, and the Moon")

color_index = 0
label_color_map = {}

for body in bodies:
    scaling = Scaling.maker(target=body)
    crater_diams = []
    morphs = []

    for d in proj_diameters:
        scaling.projectile.new_projectile()
        scaling.recompute()
        transient_diameter = scaling.projectile_to_transient(d)
        final_diameter, morphology_type = scaling.transient_to_final(transient_diameter)
        crater_diams.append(final_diameter)
        morphs.append(morphology_type)

    for morph in morphology_types:
        mask = np.array(morphs) == morph
        label = f"{body} - {morph.capitalize()}"

        ax.plot(
            proj_diameters[mask]*1e-3,
            np.array(crater_diams)[mask]*1e-3,
            markerstyle_map[morph],
            markersize=markersize,
            linestyle='None',
            label=label,
            color=color_map[body],
            markeredgewidth=0.5
        )


# Legend for planetary bodies (color only)
body_legend = [
    Line2D([0], [0], marker='s', color='w', label=body,
           markerfacecolor=color_map[body], markersize=markersize)
    for body in bodies
]

# Legend for morphology types (symbol only, black color)
morph_legend = [
    Line2D([0], [0], marker=markerstyles[i], color='k', label=morph.capitalize(),
           linestyle='None', markersize=markersize, markeredgewidth=0.5)
    for i, morph in enumerate(morphology_types)
]

# Add the legends to the plot
ax.legend(handles=body_legend + morph_legend, fontsize=9, loc='best')

ax.grid(True, which='both', linestyle='--', alpha=0.3)
plt.tight_layout()
plt.show()



import tempfile
import unittest
from pathlib import Path

import numpy as np

from cratermaker import Simulation, Surface, Target
from cratermaker.constants import _COMBINED_DATA_FILE_NAME, _EXPORT_DIR
from cratermaker.utils.general_utils import normalize_coords
from cratermaker.utils.montecarlo_utils import get_random_location

surfacetypes = Surface.available()


class TestSurface(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.target = Target(name="Moon")
        self.pix = self.target.radius / 10.0
        self.gridlevel = 4

        return

    def test_initialize_surface(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # Initializing it first should run the mesh generator
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            del surface

            # Try initializing the surface again with the same parameters. This should find the existing grid file and load it
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=False)

            # Test regridding if the parameters change
            n_face_orig = surface.uxgrid.n_face
            del surface

            surface = Surface.maker(
                simdir=simdir,
                gridlevel=self.gridlevel - 1,
                target=self.target,
                reset=False,
            )
            self.assertGreater(n_face_orig, surface.uxds.uxgrid.n_face)
            del surface

            # Test different target values
            surface = Surface.maker(
                simdir=simdir,
                gridlevel=self.gridlevel,
                target=Target(name="Mars"),
                reset=False,
            )
            del surface

            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target="Mercury", reset=False)
            del surface

            # Test bad values
            with self.assertRaises(TypeError):
                Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=1, reset=False)
            with self.assertRaises(ValueError):
                Surface.maker(
                    simdir=simdir,
                    gridlevel=self.gridlevel,
                    target="Arrakis",
                    reset=False,
                )
            with self.assertRaises(ValueError):
                Surface.maker(
                    simdir=simdir,
                    gridlevel=self.gridlevel,
                    target=Target(name="Salusa Secundus"),
                    reset=False,
                )
        return

    def test_update_elevation(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            # Test with valid elevation data
            new_elev = np.random.rand(surface.uxds.uxgrid.n_node)  # Generate random elevation data
            surface.update_elevation(new_elev)

            # Test with invalid elevation data (wrong size)
            new_elev = np.random.rand(surface.uxds.uxgrid.n_node + 1)  # Incorrect size

            # Expect ValueError for incorrect size
            with self.assertRaises(ValueError):
                surface.update_elevation(new_elev)

            surface.update_elevation(0.0, overwrite=True)

            # Check if the elevation data is set to zero
            np.testing.assert_array_equal(
                surface.uxds["node_elevation"].values,
                np.zeros(surface.uxds.uxgrid.n_node),
            )
            np.testing.assert_array_equal(
                surface.uxds["face_elevation"].values,
                np.zeros(surface.uxds.uxgrid.n_face),
            )

        return

    def test_calculate_haversine_distance(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # Example coordinates (lat/lon in radians)
            lon1, lat1 = np.radians(0), np.radians(0)  # Equator, prime meridian
            lon2, lat2 = np.radians(90), np.radians(0)  # 90 degrees East, equator
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target="Earth")

            # Known distance should be 1/4 the circumference of the Earth
            expected_distance = np.pi * surface.radius / 2
            calculated_distance = surface._calculate_distance(lon1, lat1, lon2, lat2)[0]

            # Compare the expected and calculated distances
            self.assertAlmostEqual(calculated_distance, expected_distance, places=1)

    def test_get_face_distance(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)

            location = get_random_location()[0]
            lon = location[0]
            lat = location[1]

            # Convert important points to radians
            north = (lon, lat + 90)
            south = (lon, lat - 90)
            antipode = (lon - 180, -lat)

            north = normalize_coords(north)
            south = normalize_coords(south)
            antipode = normalize_coords(antipode)

            north_idx = surface.find_nearest_face(north)
            south_idx = surface.find_nearest_face(south)
            antipode_idx = surface.find_nearest_face(antipode)

            north_distance = surface.target.radius * np.pi / 2
            south_distance = surface.target.radius * np.pi / 2
            antipode_distance = surface.target.radius * np.pi
            delta = 2 * self.pix

            # Test distances
            distances, _ = surface.calculate_face_and_node_distances(location)
            self.assertAlmostEqual(
                distances[north_idx],
                north_distance,
                delta=delta,
                msg=f"North face distance ratio: {distances[north_idx].item() / north_distance}",
            )
            self.assertAlmostEqual(
                distances[south_idx],
                south_distance,
                delta=delta,
                msg=f"South face distance ratio: {distances[south_idx].item() / south_distance}",
            )
            self.assertAlmostEqual(
                distances[antipode_idx],
                antipode_distance,
                delta=delta,
                msg=f"Antipode face distance ratio: {distances[antipode_idx].item() / antipode_distance}",
            )

    def test_get_node_distance(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)

            location = get_random_location()[0]
            lon = location[0]
            lat = location[1]

            # Convert important points to radians
            north = (lon, lat + 90)
            south = (lon, lat - 90)
            antipode = (lon + 180, -lat)

            north = normalize_coords(north)
            south = normalize_coords(south)
            antipode = normalize_coords(antipode)

            north_idx = surface.find_nearest_node(north)
            south_idx = surface.find_nearest_node(south)
            antipode_idx = surface.find_nearest_node(antipode)

            north_distance = surface.target.radius * np.pi / 2
            south_distance = surface.target.radius * np.pi / 2
            antipode_distance = surface.target.radius * np.pi
            delta = 2 * self.pix

            # Test distances
            _, node_distances = surface.calculate_face_and_node_distances(location)
            self.assertAlmostEqual(
                node_distances[north_idx],
                north_distance,
                delta=delta,
                msg=f"North node distance ratio: {node_distances[north_idx].item() / north_distance}",
            )
            self.assertAlmostEqual(
                node_distances[south_idx],
                south_distance,
                delta=delta,
                msg=f"South node distance ratio: {node_distances[south_idx].item() / south_distance}",
            )
            self.assertAlmostEqual(
                node_distances[antipode_idx],
                antipode_distance,
                delta=delta,
                msg=f"Antipode node distance ratio: {node_distances[antipode_idx].item() / antipode_distance}",
            )

    def test_calculate_initial_bearing(self):
        # Example coordinates (lat/lon in radians)
        lon1, lat1 = np.radians(0), np.radians(0)  # Equator, prime meridian
        lon2, lat2 = np.radians(90), np.radians(0)  # 90 degrees East, equator

        # The bearing from (0, 0) to (90, 0) should be 90 degrees in radians
        expected_bearing = np.radians(90)
        calculated_bearing = Surface._calculate_bearing(lon1, lat1, lon2, lat2)

        # Compare the expected and calculated bearings
        self.assertAlmostEqual(calculated_bearing, expected_bearing, places=1)

    def test_get_random_on_face(self):
        # Tests that the random location is within the face we expect
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            n_per_face = 10
            for original_face_index in range(surface.n_face):
                for _ in range(n_per_face):
                    location = surface.get_random_location_on_face(face_index=original_face_index)
                    new_face_index = surface.find_nearest_face(location.item())
                    neighbors = surface.face_face_connectivity[new_face_index].tolist()
                    neighbors.append(new_face_index)
                    self.assertIn(
                        original_face_index,
                        neighbors,
                        msg=f"Face {original_face_index} not found in neighbors of face {new_face_index}",
                    )

    def test_face_surface_values(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # Tests that the face_surface generates the correct values
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            total_area_1 = surface.uxgrid.calculate_total_face_area()
            total_area_2 = surface.face_area.sum().item()
            ratio = np.sqrt(total_area_2 / total_area_1) / self.target.radius
            self.assertAlmostEqual(ratio, 1.0, places=2)
            self.assertAlmostEqual(total_area_2 / (4 * np.pi * self.target.radius**2), 1.0, places=2)

    def test_generate_grid(self):
        gridargs = {
            "icosphere": {
                "level": self.gridlevel,
            },
            "arbitrary_resolution": {
                "pix": self.pix,
            },
            "hireslocal": {
                "pix": self.pix / 100.0,
                "local_location": (0, 0),
                "local_radius": self.pix * 2,
                "superdomain_scale_factor": 100,
            },
        }

        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            for name, args in gridargs.items():
                surface = Surface.maker(name, simdir=simdir, **args)
                self.assertTrue(Path(surface.grid_file).exists())

        return

    def test_add_data(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            test_value = 76.0
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            region = surface.extract_region(location=(0, 0), region_radius=100e3)

            for obj, uxds in zip([surface, region], [surface.uxds, region.surface.uxds], strict=False):
                surface.reset()
                obj.add_data(
                    name="test_data",
                    long_name="data for testing",
                    units="furlongs per fortnight",
                    data=test_value,
                )
                n_face = obj.n_face
                n_node = obj.n_node
                face_indices = obj.face_indices
                node_indices = obj.node_indices
                # Test that the attributes were added correctly
                self.assertIn("test_data", uxds)
                self.assertEqual(uxds["test_data"].long_name, "data for testing")
                self.assertEqual(uxds["test_data"].units, "furlongs per fortnight")
                np.testing.assert_array_equal(
                    uxds["test_data"].sel(n_face=face_indices).values,
                    np.full(n_face, test_value),
                )

                obj.add_data(
                    name="test_data",
                    data=-test_value,
                    long_name="this should be ignored",
                    units="slug angstrom^2 / fortnight^2",
                    overwrite=True,
                )
                # test that the attributes were ignored but the data was added
                self.assertIn("test_data", uxds)
                self.assertEqual(uxds["test_data"].long_name, "data for testing")
                self.assertEqual(uxds["test_data"].units, "furlongs per fortnight")
                np.testing.assert_array_equal(
                    uxds["test_data"].sel(n_face=face_indices).values,
                    np.full(n_face, -test_value),
                )

                # Add scalar node data
                obj.add_data(name="scalar_node", data=test_value, isfacedata=False)
                self.assertIn("scalar_node", uxds)
                np.testing.assert_array_equal(
                    uxds["scalar_node"].sel(n_node=node_indices).values,
                    np.full(n_node, test_value),
                )

                # Add array face data
                data_array = np.arange(n_face)
                obj.add_data(name="array_face", data=data_array)
                np.testing.assert_array_equal(uxds["array_face"].sel(n_face=face_indices).values, data_array)

                # Overwrite behavior
                obj.add_data(name="array_face", data=np.ones(n_face), overwrite=False)
                np.testing.assert_array_equal(
                    uxds["array_face"].sel(n_face=face_indices).values,
                    data_array + 1,
                )

                obj.add_data(name="array_face", data=np.zeros(n_face), overwrite=True)
                np.testing.assert_array_equal(
                    uxds["array_face"].sel(n_face=face_indices).values,
                    np.zeros(n_face),
                )

    def test_diffusion(self):
        def initial_elevation(r, h0, sigma):
            return h0 * np.exp(-(r**2) / (2 * sigma**2))

        def analytical_elevation(r, h0, sigma, kappa):
            variance_t = sigma**2 + 2 * kappa
            return h0 * (sigma**2 / variance_t) * np.exp(-(r**2) / (2 * variance_t))

        local_radius = 2000.0
        pix = 10.0
        local_location = (0, 0)
        h0 = 100.0
        sigma = 200.0
        kdiff = 20000.0

        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            sim = Simulation(simdir=simdir, surface="hireslocal", local_location=local_location, pix=pix, local_radius=local_radius)

            region_view = sim.surface.local
            rvals = region_view.face_distance

            # Initial elevation setup
            h_initial = initial_elevation(rvals, h0, sigma)
            region_view.update_elevation(h_initial)

            # Compute the analytical solution result after
            h_analytical = analytical_elevation(rvals, h0, sigma, kdiff)
            hnorm = h_analytical.max()
            h_analytical /= hnorm  # Normalize the elevations so that the maximum expected value is 1.0
            region_view.apply_diffusion(kdiff)
            h_numerical = region_view.face_elevation / hnorm
            np.testing.assert_array_almost_equal(
                h_analytical, h_numerical, decimal=2, err_msg="Diffusion did not match analytical solution"
            )

    def test_export_vtk(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel)
            # Test with default parameters
            default_out_dir = Path(surface.simdir) / _EXPORT_DIR
            expected_files = ["surface000000.vtp", "grid.vtp"]
            surface.export("vtp")
            self.assertTrue(Path(default_out_dir).is_dir())
            for f in expected_files:
                self.assertTrue(Path(default_out_dir / f).exists())


if __name__ == "__main__":
    unittest.main()
